import shutil
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np
import tensorflow as tf

from ..interface import ProblemBase
from ..registry import cfg_serializable, get_class


@dataclass
class ClassificationConfig:
    model: Any
    model_class: str
    optimizer: Any
    optimizer_class: str = "OptimizerFactory"
    # We treat binary classification problems as a special case, because for binary
    # problems the model can return just one logit, which is the logit for class 1.
    # The logit for class 0 is implicitly set to 0.0.
    binary_loss: bool = False
    # We apply weight decay by summing the squares of all trainable variables and
    # multiplying them by `weight_decay`. We are ignoring Keras weight regularizers
    # and the automatically generated model losses.
    weight_decay: float = 0.0
    mixed_precision: bool = False
    # When saving the model we may want to use a different dtype for model inputs. E.g.,
    # for images, `uint8` is a natural choice. In particular if the saved model is
    # deployed via TF serving, `uint8` input reduces the network payload, even though
    # the first thing the model does is cast everything to `float32`.
    save_input_dtype: str = "float32"


@cfg_serializable
class ClassificationProblem(ProblemBase):
    cfg_class = ClassificationConfig

    def __init__(self, cfg: ClassificationConfig, timekeeping):
        self.cfg = cfg
        self.timekeeping = timekeeping

        # Setting global state before building model
        if cfg.mixed_precision:
            tf.keras.mixed_precision.set_global_policy("mixed_float16")

        # Building the model
        model, preprocess = get_class(cfg.model_class)(cfg=cfg.model)()
        self.model = model
        self.preprocess = preprocess

        # Training metrics
        self.avg_ce_loss = tf.keras.metrics.Mean(dtype=tf.float32)
        self.avg_reg_loss = tf.keras.metrics.Mean(dtype=tf.float32)
        self.avg_loss = tf.keras.metrics.Mean(dtype=tf.float32)
        self.avg_acc = tf.keras.metrics.Accuracy(dtype=tf.float32)

        # Optimizer
        self.optimizer = get_class(cfg.optimizer_class)(
            cfg=cfg.optimizer,
            timekeeping=timekeeping,
            mixed_precision=cfg.mixed_precision,
        )()

    def ckpt_variables(self, model_only: bool = False):
        """Return dictionary with all variables that need to be added to checkpoint."""
        variables = {"model": self.model}
        if not model_only:
            variables["avg_ce_loss"] = self.avg_ce_loss
            variables["avg_reg_loss"] = self.avg_reg_loss
            variables["avg_loss"] = self.avg_loss
            variables["avg_acc"] = self.avg_acc
            variables["optimizer"] = self.optimizer
        return variables

    def start_epoch(self):
        """Called at the beginning of an epoch. Used to reset moving averages."""
        self.avg_ce_loss.reset_states()
        self.avg_reg_loss.reset_states()
        self.avg_loss.reset_states()
        self.avg_acc.reset_states()

    def train_step(self, data, it):
        """Perform one step of training."""
        img, labels = data
        ce_loss, reg_loss, preds = self.train_step_inner(img, labels)

        self.avg_ce_loss.update_state(ce_loss)
        self.avg_reg_loss.update_state(reg_loss)
        self.avg_loss.update_state(reg_loss + ce_loss)
        self.avg_acc.update_state(preds, labels)

        logs = {
            "train/ce_loss": self.avg_ce_loss.result().numpy(),
            "train/reg_loss": self.avg_reg_loss.result().numpy(),
            "train/loss": self.avg_loss.result().numpy(),
            "train/acc": self.avg_acc.result().numpy(),
        }
        return logs["train/loss"], logs

    @tf.function
    def train_step_inner(self, img, labels):
        img = self.preprocess(img)

        with tf.GradientTape() as tape:
            logits = self.logits(img, training=True)
            # Regardless of mixed precision or not, we compute the loss in float32
            logits = tf.cast(logits, tf.float32)
            ce_loss = self.softmax_loss(logits, labels)

            # Weight decay
            # TODO: Exclude certain variables from weight decay based on model cfg
            reg_loss = 0.0
            if self.cfg.weight_decay != 0.0:
                for weight in self.model.trainable_variables:
                    reg_loss += tf.reduce_sum(tf.square(weight))
                reg_loss *= self.cfg.weight_decay

            # Total loss
            loss = ce_loss + reg_loss
            if self.cfg.mixed_precision:
                loss = self.optimizer.get_scaled_loss(loss)

        grads = tape.gradient(loss, self.model.trainable_variables)
        if self.cfg.mixed_precision:
            grads = self.optimizer.get_unscaled_gradients(grads)
        self.optimizer.apply_gradients(zip(grads, self.model.trainable_variables))

        preds = self.predict(logits)
        return ce_loss, reg_loss, preds

    def logits(self, img, training):
        logits = self.model(img, training=training)
        logits = tf.cast(logits, tf.float32)
        return logits

    def predict(self, logits):
        if self.cfg.binary_loss:
            logits = tf.concat([tf.zeros_like(logits), logits], axis=-1)
        preds = tf.argmax(logits, axis=-1)
        return preds

    def softmax_loss(self, logits, labels):
        if self.cfg.binary_loss:
            loss = tf.nn.sigmoid_cross_entropy_with_logits(
                labels=tf.cast(labels, tf.float32), logits=logits[:, 0]
            )
        else:
            loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
                labels=labels, logits=logits
            )
        loss = tf.reduce_mean(loss)
        return loss

    def logits_and_labels(self, dataset):
        """
        Returns logits and labels from dataset. The dataset must be finite.

        Returns:
            np.ndarrays for logits and labels of shape (N, nb_classes) and (N,)
        """

        def _inference(_img):
            _img = self.preprocess(_img)
            _logits = self.model(_img, training=False)
            _logits = tf.cast(_logits, tf.float32)
            if self.cfg.binary_loss:
                # In the binary case the model can return logits only for class 1.
                # The logit for class 0 is assumed to be 0.0.
                _logits = _logits[..., 0]
                _logits = tf.stack([tf.zeros_like(_logits), _logits], axis=-1)
            # Normalize logits to have sum=0. While not necessary to compute
            # accuracy, this becomes important if we want to compare decision
            # thresholds across epochs and training runs.
            _logits = _logits - tf.reduce_mean(_logits, axis=-1, keepdims=True)
            return _logits

        labels, logits = [], []
        for img_batch, labels_batch in dataset:
            logits_batch = _inference(img_batch)
            logits.append(logits_batch.numpy())
            labels.append(labels_batch.numpy())
        labels = np.concatenate(labels, axis=0)
        logits = np.concatenate(logits, axis=0)
        return logits, labels

    def validation(self, dataset):
        """
        Function performs validation on a dataset and returns a dictionary of metrics.
        """
        logits, labels = self.logits_and_labels(dataset)
        preds = tf.argmax(logits, axis=-1).numpy()

        # This is work in progress. For now, we measure only accuracy. Later, we should
        # add top-5 accuracy, etc.
        acc = np.sum(preds == labels) / len(labels)

        logs = {"val/acc": acc}
        return logs["val/acc"], logs

    def save_model(self, save_dir):
        """Save models ready for inference."""
        save_dir = Path(save_dir)

        # We need to set policy to float32 for saving, otherwise we save models that
        # perform inference with float16, which is extremely slow on CPUs
        old_policy = tf.keras.mixed_precision.global_policy()
        tf.keras.mixed_precision.set_global_policy("float32")
        # After changing the policy, we need to create a new model using the policy
        model_factory = get_class(self.cfg.model_class)(cfg=self.cfg.model)
        save_model, save_preprocess = model_factory()
        save_model.set_weights(self.model.get_weights())

        # Now build the full inference model including preprocessing and logit layer
        inputs = tf.keras.layers.Input(
            shape=model_factory.tf_input_shape,
            batch_size=None,
            dtype=self.cfg.save_input_dtype,
            name="input",
        )
        img = tf.cast(inputs, tf.float32)
        img = save_preprocess(img)
        logits = save_model(img, training=False)
        if self.cfg.binary_loss:
            # In the binary case the model can return logits only for class 1.
            # The logit for class 0 is assumed to be 0.0.
            logits = logits[..., 0]
            logits = tf.stack([tf.zeros_like(logits), logits], axis=-1)
        # Normalize logits to have sum=0.
        logits = logits - tf.reduce_mean(logits, axis=-1, keepdims=True)
        # So output layer has the right name
        logits = tf.keras.layers.Activation("linear", name="logits")(logits)
        inference_model = tf.keras.Model(inputs, logits)

        model_dir = save_dir / "model"
        with tempfile.TemporaryDirectory() as tmpdir:
            # If `save_dir` points to a network file system or S3FS, sometimes TF saving
            # can be very slow. It is faster to save to a temporary directory first and
            # copying data across.
            local_dir = Path(tmpdir) / "model"
            tf.saved_model.save(inference_model, str(local_dir))
            # TODO: Add support for S3 paths here...
            shutil.copytree(str(local_dir), str(model_dir), dirs_exist_ok=True)

        # Restore original float policy
        tf.keras.mixed_precision.set_global_policy(old_policy)



"""
TensorFlow implementation of the Visual Transformer

Based on timm/models/visual_transformer.py by Ross Wightman.
Based on transformers/models/vit by HuggingFace

Copyright 2021 Martins Bruveris
"""
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import tensorflow as tf

from tfimm.layers import (
    MLP,
    DropPath,
    PatchEmbeddings,
    interpolate_pos_embeddings,
    norm_layer_factory,
)
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import (
    IMAGENET_DEFAULT_MEAN,
    IMAGENET_DEFAULT_STD,
    IMAGENET_INCEPTION_MEAN,
    IMAGENET_INCEPTION_STD,
)

from .resnetv2 import ResNetV2, ResNetV2Config, ResNetV2Stem

# model_registry will add each entrypoint fn to this
__all__ = ["ViT", "ViTBlock", "ViTConfig"]


@dataclass
class ViTConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_layer: str = "patch_embeddings"
    patch_nb_blocks: tuple = ()
    patch_size: int = 16
    embed_dim: int = 768
    nb_blocks: int = 12
    nb_heads: int = 12
    mlp_ratio: float = 4.0
    qkv_bias: bool = True
    representation_size: Optional[int] = None
    distilled: bool = False
    # Regularization
    drop_rate: float = 0.0
    attn_drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Other parameters
    norm_layer: str = "layer_norm_eps_1e-6"
    act_layer: str = "gelu"
    # Parameters for inference
    interpolate_input: bool = False
    crop_pct: float = 0.875
    interpolation: str = "bicubic"
    mean: Tuple[float, float, float] = IMAGENET_INCEPTION_MEAN
    std: Tuple[float, float, float] = IMAGENET_INCEPTION_STD
    first_conv: str = "patch_embed/proj"
    # DeiT models have two classifier heads, one for distillation
    classifier: Union[str, Tuple[str, str]] = "head"

    """
    Args:
        nb_classes: Number of classes for classification head
        in_channels: Number of input channels
        input_size: Input image size
        patch_layer: Layer used to transform image to patches. Possible values are
            `patch_embeddings` and `hybrid_embeddings`.
        patch_nb_blocks: When `patch_layer="hybrid_embeddings`, this is the number of
            residual blocks in each stage. Set to `()` to use only the stem.
        patch_size: Patch size; Image size must be multiple of patch size. For hybrid
            embedding layer, this patch size is applied after the convolutional layers.
        embed_dim: Embedding dimension
        nb_blocks: Depth of transformer (number of encoder blocks)
        nb_heads: Number of self-attention heads
        mlp_ratio: Ratio of mlp hidden dim to embedding dim
        qkv_bias: Enable bias for qkv if True
        representation_size: Enable and set representation layer (pre-logits) to this
            value if set
        distilled: Model includes a distillation token and head as in DeiT models
        drop_rate: Dropout rate
        attn_drop_rate: Attention dropout rate
        drop_path_rate: Dropout rate for stochastic depth
        norm_layer: Normalization layer
        act_layer: Activation function
    """

    @property
    def nb_tokens(self) -> int:
        """Number of special tokens"""
        return 2 if self.distilled else 1

    @property
    def grid_size(self) -> Tuple[int, int]:
        grid_size = (
            self.input_size[0] // self.patch_size,
            self.input_size[1] // self.patch_size,
        )
        if self.patch_layer == "hybrid_embeddings":
            # 2 reductions in the stem, 1 reduction in each stage except the first one
            reductions = 2 + max(len(self.patch_nb_blocks) - 1, 0)
            stride = 2**reductions
            grid_size = (grid_size[0] // stride, grid_size[1] // stride)
        return grid_size

    @property
    def nb_patches(self) -> int:
        """Number of patches without class and distillation tokens."""
        return self.grid_size[0] * self.grid_size[1]

    @property
    def transform_weights(self):
        return {"pos_embed": ViT.transform_pos_embed}


class ViTMultiHeadAttention(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        qkv_bias: bool,
        drop_rate: float,
        attn_drop_rate: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads
        self.qkv_bias = qkv_bias
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate

        head_dim = embed_dim // nb_heads
        self.scale = head_dim**-0.5

        self.qkv = tf.keras.layers.Dense(
            units=3 * embed_dim, use_bias=qkv_bias, name="qkv"
        )
        self.attn_drop = tf.keras.layers.Dropout(rate=attn_drop_rate)
        self.proj = tf.keras.layers.Dense(units=embed_dim, name="proj")
        self.proj_drop = tf.keras.layers.Dropout(rate=drop_rate)

    def call(self, x, training=False, return_features=False):
        features = OrderedDict()

        # B (batch size), N (sequence length), D (embedding dimension),
        # H (number of heads)
        batch_size, seq_length = tf.unstack(tf.shape(x)[:2])
        qkv = self.qkv(x)  # (B, N, 3*D)
        qkv = tf.reshape(qkv, (batch_size, seq_length, 3, self.nb_heads, -1))
        qkv = tf.transpose(qkv, (2, 0, 3, 1, 4))  # (3, B, H, N, D/H)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = self.scale * tf.linalg.matmul(q, k, transpose_b=True)  # (B, H, N, N)
        attn = tf.nn.softmax(attn, axis=-1)  # (B, H, N, N)
        attn = self.attn_drop(attn, training=training)
        features["attn"] = attn

        x = tf.linalg.matmul(attn, v)  # (B, H, N, D/H)
        x = tf.transpose(x, (0, 2, 1, 3))  # (B, N, H, D/H)
        x = tf.reshape(x, (batch_size, seq_length, -1))  # (B, N, D)

        x = self.proj(x)
        x = self.proj_drop(x, training=training)
        return (x, features) if return_features else x


class ViTBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        drop_rate: float,
        attn_drop_rate: float,
        drop_path_rate: float,
        norm_layer: str,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.drop_path_rate = drop_path_rate
        self.norm_layer = norm_layer
        self.act_layer = act_layer
        norm_layer = norm_layer_factory(norm_layer)

        self.norm1 = norm_layer(name="norm1")
        self.attn = ViTMultiHeadAttention(
            embed_dim=embed_dim,
            nb_heads=nb_heads,
            qkv_bias=qkv_bias,
            drop_rate=drop_rate,
            attn_drop_rate=attn_drop_rate,
            name="attn",
        )
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.norm2 = norm_layer(name="norm2")
        self.mlp = MLP(
            hidden_dim=int(embed_dim * mlp_ratio),
            embed_dim=embed_dim,
            drop_rate=drop_rate,
            act_layer=act_layer,
            name="mlp",
        )

    def call(self, x, training=False, return_features=False):
        features = OrderedDict()
        shortcut = x
        x = self.norm1(x, training=training)
        x = self.attn(x, training=training, return_features=return_features)
        if return_features:
            x, mha_features = x
            features["attn"] = mha_features["attn"]
        x = self.drop_path(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp(x, training=training)
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return (x, features) if return_features else x


class HybridEmbeddings(tf.keras.layers.Layer):
    """
    CNN feature map embedding

    Extract feature map from CNN, flatten, project to embedding dim.
    """

    def __init__(
        self,
        in_channels: int,
        input_size: tuple,
        nb_blocks: tuple,
        patch_size: int,
        embed_dim: int,
        drop_path_rate: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if nb_blocks == ():
            self.backbone = ResNetV2Stem(
                stem_type="same",
                stem_width=64,
                conv_padding="same",
                preact=False,
                act_layer="relu",
                norm_layer="group_norm",
                name="backbone",
            )
        else:
            backbone_cfg = ResNetV2Config(
                nb_classes=0,
                in_channels=in_channels,
                input_size=input_size,
                nb_blocks=nb_blocks,
                preact=False,
                stem_type="same",
                global_pool="",
                conv_padding="same",
                drop_path_rate=drop_path_rate,
            )
            self.backbone = ResNetV2(backbone_cfg, name="backbone")

        self.projection = tf.keras.layers.Conv2D(
            filters=embed_dim,
            kernel_size=patch_size,
            strides=patch_size,
            use_bias=True,
            name="proj",
        )

    def call(self, x, training=False, return_shape=False):
        x = self.backbone(x, training=training)
        x = self.projection(x)

        # Change the 2D spatial dimensions to a single temporal dimension.
        batch_size, height, width = tf.unstack(tf.shape(x)[:3])
        x = tf.reshape(tensor=x, shape=(batch_size, height * width, -1))
        return (x, (height, width)) if return_shape else x


@keras_serializable
class ViT(tf.keras.Model):
    cfg_class = ViTConfig

    def __init__(self, cfg: ViTConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.nb_features = cfg.embed_dim  # For consistency with other models
        self.norm_layer = norm_layer_factory(cfg.norm_layer)
        self.cfg = cfg

        if cfg.patch_layer == "patch_embeddings":
            self.patch_embed = PatchEmbeddings(
                patch_size=cfg.patch_size,
                embed_dim=cfg.embed_dim,
                norm_layer="",  # ViT does not use normalization in patch embeddings
                name="patch_embed",
            )
        elif cfg.patch_layer == "hybrid_embeddings":
            self.patch_embed = HybridEmbeddings(
                in_channels=cfg.in_channels,
                input_size=cfg.input_size,
                nb_blocks=cfg.patch_nb_blocks,
                patch_size=cfg.patch_size,
                embed_dim=cfg.embed_dim,
                drop_path_rate=cfg.drop_path_rate,
                name="patch_embed",
            )
        else:
            raise ValueError(f"Unknown patch layer: {cfg.patch_layer}.")
        self.cls_token = None
        self.dist_token = None
        self.pos_embed = None
        self.pos_drop = tf.keras.layers.Dropout(rate=cfg.drop_rate)

        self.blocks = [
            ViTBlock(
                embed_dim=cfg.embed_dim,
                nb_heads=cfg.nb_heads,
                mlp_ratio=cfg.mlp_ratio,
                qkv_bias=cfg.qkv_bias,
                drop_rate=cfg.drop_rate,
                attn_drop_rate=cfg.attn_drop_rate,
                drop_path_rate=cfg.drop_path_rate,
                norm_layer=cfg.norm_layer,
                act_layer=cfg.act_layer,
                name=f"blocks/{j}",
            )
            for j in range(cfg.nb_blocks)
        ]
        self.norm = self.norm_layer(name="norm")

        # Some models have a representation layer on top of cls token
        if cfg.representation_size:
            if cfg.distilled:
                raise ValueError(
                    "Cannot combine distillation token and a representation layer."
                )
            self.nb_features = cfg.representation_size
            self.pre_logits = tf.keras.layers.Dense(
                units=cfg.representation_size, activation="tanh", name="pre_logits/fc"
            )
        else:
            self.pre_logits = None

        # Classifier head(s)
        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )
        if cfg.distilled:
            self.head_dist = (
                tf.keras.layers.Dense(units=cfg.nb_classes, name="head_dist")
                if cfg.nb_classes > 0
                else tf.keras.layers.Activation("linear")  # Identity layer
            )
        else:
            self.head_dist = None

    def build(self, input_shape):
        self.cls_token = self.add_weight(
            shape=(1, 1, self.cfg.embed_dim),
            initializer="zeros",
            trainable=True,
            name="cls_token",
        )
        self.dist_token = (
            self.add_weight(
                shape=(1, 1, self.cfg.embed_dim),
                initializer="zeros",
                trainable=True,
                name="dist_token",
            )
            if self.cfg.distilled
            else None
        )
        self.pos_embed = self.add_weight(
            shape=(1, self.cfg.nb_patches + self.cfg.nb_tokens, self.cfg.embed_dim),
            initializer="zeros",
            trainable=True,
            name="pos_embed",
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        """
        Names of features, returned when calling ``call`` with ``return_features=True``.
        """
        _, features = self(self.dummy_inputs, return_features=True)
        return list(features.keys())

    def transform_pos_embed(self, src_weights, target_cfg: ViTConfig):
        return interpolate_pos_embeddings(
            pos_embed=self.pos_embed,
            src_grid_size=self.cfg.grid_size,
            tgt_grid_size=target_cfg.grid_size,
            nb_tokens=self.cfg.nb_tokens,
        )

    def forward_features(self, x, training=False, return_features=False):
        features = OrderedDict()
        batch_size = tf.shape(x)[0]

        x, grid_size = self.patch_embed(x, return_shape=True)
        cls_token = tf.repeat(self.cls_token, repeats=batch_size, axis=0)
        if not self.cfg.distilled:
            x = tf.concat((cls_token, x), axis=1)
        else:
            dist_token = tf.repeat(self.dist_token, repeats=batch_size, axis=0)
            x = tf.concat((cls_token, dist_token, x), axis=1)
        if not self.cfg.interpolate_input:
            x = x + self.pos_embed
        else:
            pos_embed = interpolate_pos_embeddings(
                self.pos_embed,
                src_grid_size=self.cfg.grid_size,
                tgt_grid_size=grid_size,
                nb_tokens=self.cfg.nb_tokens,
            )
            x = x + pos_embed
        x = self.pos_drop(x, training=training)
        features["patch_embedding"] = x

        for j, block in enumerate(self.blocks):
            x = block(x, training=training, return_features=return_features)
            if return_features:
                x, block_features = x
                features[f"block_{j}/attn"] = block_features["attn"]
            features[f"block_{j}"] = x
        x = self.norm(x, training=training)
        features["features_all"] = x

        if self.cfg.distilled:
            # Here we diverge from timm and return both outputs as one tensor. That way
            # all models always have one output by default
            x = x[:, :2]
        elif self.cfg.representation_size:
            x = self.pre_logits(x[:, 0])
        else:
            x = x[:, 0]
        features["features"] = x
        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        if not self.cfg.distilled:
            x = self.head(x)
        else:
            y = self.head(x[:, 0])
            y_dist = self.head_dist(x[:, 1])
            x = tf.stack((y, y_dist), axis=1)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def vit_tiny_patch16_224():
    """ViT-Tiny (Vit-Ti/16)"""
    cfg = ViTConfig(
        name="vit_tiny_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
    )
    return ViT, cfg


@register_model
def vit_tiny_patch16_384():
    """ViT-Tiny (Vit-Ti/16) @ 384x384."""
    cfg = ViTConfig(
        name="vit_tiny_patch16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_small_patch32_224():
    """ViT-Small (ViT-S/32)"""
    cfg = ViTConfig(
        name="vit_small_patch32_224",
        url="[timm]",
        patch_size=32,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
    )
    return ViT, cfg


@register_model
def vit_small_patch32_384():
    """ViT-Small (ViT-S/32) at 384x384."""
    cfg = ViTConfig(
        name="vit_small_patch32_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=32,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_small_patch16_224():
    """ViT-Small (ViT-S/16)"""
    cfg = ViTConfig(
        name="vit_small_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
    )
    return ViT, cfg


@register_model
def vit_small_patch16_384():
    """ViT-Small (ViT-S/16)"""
    cfg = ViTConfig(
        name="vit_small_patch16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_base_patch32_224():
    """
    ViT-Base (ViT-B/32) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_patch32_224",
        url="[timm]",
        patch_size=32,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_base_patch32_384():
    """
    ViT-Base model (ViT-B/32) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 384x384, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_patch32_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=32,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_base_patch16_224():
    """
    ViT-Base (ViT-B/16) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 224x224, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_base_patch16_384():
    """
    ViT-Base model (ViT-B/16) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 384x384, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_patch16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_base_patch8_224():
    """
    ViT-Base (ViT-B/8) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 224x224, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_patch8_224",
        url="[timm]",
        patch_size=8,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_large_patch32_224():
    """
    ViT-Large model (ViT-L/32) from original paper (https://arxiv.org/abs/2010.11929).
    No pretrained weights.
    """
    cfg = ViTConfig(
        name="vit_large_patch32_224",
        url="[timm]",
        patch_size=32,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
    )
    return ViT, cfg


@register_model
def vit_large_patch32_384():
    """
    ViT-Large model (ViT-L/32) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 384x384, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_large_patch32_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=32,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_large_patch16_224():
    """
    ViT-Large model (ViT-L/32) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 224x224, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_large_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
    )
    return ViT, cfg


@register_model
def vit_large_patch16_384():
    """
    ViT-Large model (ViT-L/16) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 384x384, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_large_patch16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
        crop_pct=1.0,
    )
    return ViT, cfg


@register_model
def vit_base_patch32_sam_224():
    """
    ViT-Base (ViT-B/32) w/ SAM pretrained weights.
    Paper: https://arxiv.org/abs/2106.01548
    """
    cfg = ViTConfig(
        name="vit_base_patch32_sam_224",
        url="[timm]",
        patch_size=32,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_base_patch16_sam_224():
    """
    ViT-Base (ViT-B/16) w/ SAM pretrained weights.
    Paper: https://arxiv.org/abs/2106.01548
    """
    cfg = ViTConfig(
        name="vit_base_patch16_sam_224",
        url="[timm]",
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_tiny_patch16_224_in21k():
    """
    ViT-Tiny (Vit-Ti/16). ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_tiny_patch16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=16,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
    )
    return ViT, cfg


@register_model
def vit_small_patch32_224_in21k():
    """
    ViT-Small (ViT-S/16) ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_small_patch32_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=32,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
    )
    return ViT, cfg


@register_model
def vit_small_patch16_224_in21k():
    """
    ViT-Small (ViT-S/16) ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_small_patch16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
    )
    return ViT, cfg


@register_model
def vit_base_patch32_224_in21k():
    """
    ViT-Base model (ViT-B/32) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_base_patch32_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=32,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_base_patch16_224_in21k():
    """
    ViT-Base model (ViT-B/16) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_base_patch16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_base_patch8_224_in21k():
    """ViT-Base model (ViT-B/8) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_base_patch8_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=8,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
    )
    return ViT, cfg


@register_model
def vit_large_patch32_224_in21k():
    """
    ViT-Large model (ViT-L/32) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a representation layer but the 21k classifier head is zero'd
    out in original weights.
    """
    cfg = ViTConfig(
        name="vit_large_patch32_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=32,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
        representation_size=1024,
    )
    return ViT, cfg


@register_model
def vit_large_patch16_224_in21k():
    """
    ViT-Large model (ViT-L/16) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a valid 21k classifier head and no representation layer.
    """
    cfg = ViTConfig(
        name="vit_large_patch16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=16,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
    )
    return ViT, cfg


@register_model
def vit_huge_patch14_224_in21k():
    """
    ViT-Huge model (ViT-H/14) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    Note: This model has a representation layer but the 21k classifier head is zero'd
    out in original weights.
    """
    cfg = ViTConfig(
        name="vit_huge_patch14_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=14,
        embed_dim=1280,
        nb_blocks=32,
        nb_heads=16,
        representation_size=1280,
    )
    return ViT, cfg


@register_model
def deit_tiny_patch16_224():
    """
    DeiT-tiny model @ 224x224 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_tiny_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return ViT, cfg


@register_model
def deit_small_patch16_224():
    """
    DeiT-small model @ 224x224 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_small_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return ViT, cfg


@register_model
def deit_base_patch16_224():
    """
    DeiT base model @ 224x224 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_base_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return ViT, cfg


@register_model
def deit_base_patch16_384():
    """
    DeiT base model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_base_patch16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        crop_pct=1.0,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return ViT, cfg


@register_model
def deit_tiny_distilled_patch16_224():
    """
    DeiT-tiny distilled model @ 224x224 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_tiny_distilled_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
        distilled=True,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    )
    return ViT, cfg


@register_model
def deit_small_distilled_patch16_224():
    """
    DeiT-small distilled model @ 224x224 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_small_distilled_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        distilled=True,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    )
    return ViT, cfg


@register_model
def deit_base_distilled_patch16_224():
    """
    DeiT-base distilled model @ 224x224 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_base_distilled_patch16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        distilled=True,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    )
    return ViT, cfg


@register_model
def deit_base_distilled_patch16_384():
    """
    DeiT-base distilled model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    cfg = ViTConfig(
        name="deit_base_distilled_patch16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        distilled=True,
        crop_pct=1.0,
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    )
    return ViT, cfg


@register_model
def vit_base_patch16_224_miil_in21k():
    """ViT-Base (ViT-B/16) from original paper (https://arxiv.org/abs/2010.11929).
    Weights taken from: https://github.com/Alibaba-MIIL/ImageNet21K
    See paper: https://arxiv.org/pdf/2104.10972v4.pdf
    """
    cfg = ViTConfig(
        name="vit_base_patch16_224_miil_in21k",
        url="[timm]",
        nb_classes=11221,
        input_size=(224, 224),
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        qkv_bias=False,
        crop_pct=0.875,
        interpolation="bilinear",
        mean=(0.0, 0.0, 0.0),
        std=(0.0, 0.0, 0.0),
    )
    return ViT, cfg


@register_model
def vit_base_patch16_224_miil():
    """
    ViT-Base (ViT-B/16) from original paper (https://arxiv.org/abs/2010.11929).
    Weights taken from: https://github.com/Alibaba-MIIL/ImageNet21K
    See paper: https://arxiv.org/pdf/2104.10972v4.pdf
    """
    cfg = ViTConfig(
        name="vit_base_patch16_224_miil",
        url="[timm]",
        input_size=(224, 224),
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        qkv_bias=False,
        crop_pct=0.875,
        interpolation="bilinear",
        mean=(0.0, 0.0, 0.0),
        std=(0.0, 0.0, 0.0),
    )
    return ViT, cfg



from .config import (  # noqa: F401
    deep_to_flat,
    dump_config,
    flat_to_deep,
    parse_args,
    pprint,
    to_dict_format,
)
from .datasets import TFDSConfig, TFDSWrapper  # noqa: F401
from .interface import ProblemBase  # noqa: F401
from .model import (  # noqa: F401
    EmbeddingModelConfig,
    EmbeddingModelFactory,
    ModelConfig,
    ModelFactory,
    SavedModel,
    SavedModelConfig,
)
from .optimizers.optimizer import OptimizerConfig, OptimizerFactory  # noqa: F401
from .optimizers.schedules import (  # noqa: F401
    LRConstConfig,
    LRConstFactory,
    LRCosineDecayConfig,
    LRCosineDecayFactory,
    LRExpDecayConfig,
    LRExponentialDecayFactory,
    LRMultiStepsConfig,
    LRMultiStepsFactory,
)
from .problems import (  # noqa: F401
    ClassificationConfig,
    ClassificationProblem,
    DistillationConfig,
    DistillationProblem,
)
from .registry import cfg_serializable, get_cfg_class, get_class  # noqa: F401
from .timekeeping import Timekeeping  # noqa: F401
from .train import ExperimentConfig, run  # noqa: F401
from .trainer import SingleGPUTrainer, TrainerConfig  # noqa: F401
from .utils import (  # noqa: F401
    collect_files_with_suffix,
    collect_tfrecord_files,
    setup_logging,
)



from typing import Tuple

import tensorflow as tf

from tfimm.layers import act_layer_factory, norm_layer_factory


class MaskDecoder(tf.keras.Model):
    def __init__(
        self,
        *,
        transformer: tf.keras.Model,
        embed_dim: int,
        nb_multimask_outputs: int,
        act_layer: str,
        iou_head_depth: int,
        iou_head_hidden_dim: int,
        **kwargs,
    ) -> None:
        """
        Predicts masks given an image and prompt embeddings, using a tranformer
        architecture.

        Args:
            transformer: Transformer model used to predict masks.
            embed_dim: Dimension of image and prompt embeddings.
            nb_multimask_outputs: The number of masks to predict with disambiguating
                masks.
            act_layer: Activation layer.
            iou_head_depth: The depth of MLP used to predict mask quality.
            iou_head_hidden_dim: Hidden dimension of MLP used to predict mask quality.
        """
        super().__init__(**kwargs)
        self.transformer = transformer
        self.embed_dim = embed_dim
        self.nb_multimask_outputs = nb_multimask_outputs
        self.act_layer = act_layer
        self.iou_head_depth = iou_head_depth
        self.iou_head_hidden_dim = iou_head_hidden_dim

        self.iou_token = None
        self.mask_tokens = None
        self.nb_mask_tokens = self.nb_multimask_outputs + 1

        self.output_upscaling = OutputUpscaling(
            embed_dim=self.embed_dim, act_layer=self.act_layer, name="output_upscaling"
        )
        self.output_hypernetworks_mlps = [
            MLP(
                hidden_dim=self.embed_dim,
                output_dim=self.embed_dim // 8,
                nb_layers=3,
                name=f"output_hypernetworks_mlps/{j}",
            )
            for j in range(self.nb_mask_tokens)
        ]
        self.iou_prediction_head = MLP(
            hidden_dim=self.iou_head_hidden_dim,
            output_dim=self.nb_mask_tokens,
            nb_layers=self.iou_head_depth,
            name="iou_prediction_head",
        )

    def build(self, input_shape):
        self.iou_token = self.add_weight(
            shape=(1, self.embed_dim),
            initializer=tf.keras.initializers.RandomNormal(),
            trainable=True,
            name="iou_token/weight",
        )
        self.mask_tokens = self.add_weight(
            shape=(self.nb_mask_tokens, self.embed_dim),
            initializer=tf.keras.initializers.RandomNormal(),
            trainable=True,
            name="mask_tokens/weight",
        )

    def call(self, inputs, training=False, multimask_output: bool = False):
        """
        Predict masks given image and prompt embeddings.

        Args:
            inputs: Dictionary with the following entries.
                image_embeddings: Image embeddings of shape (N, H, W, C), where (H, W)
                    is the grid size (usually image size / 16) and C = embed_dim is the
                    dimension of image and prompt embeddings.
                image_pe: Positional encodings for the image, same shape as image
                    embeddings, i.e., (N, H, W, C).
                sparse_embeddings: Sparse prompt embeddings, shape (N, L, C), where
                    L can be 0 if no sparse embeddings are provided.
                dense_embeddings: Dense prompt embeddings, shape (N, H, W, C).
            training: Training of inference phase?
            multimask_output: If true, we return multiple predicted segmentation masks,
                the number determined by `nb_multimask_outputs`.

        Returns:
            An (N, K, H', W') tensor of predicted segmentation masks, where K is either
                1 or nb_multimask_outputs. (H', W') is the spatial resolution of the
                segmentation mask (usually image_size / 4).
            An (N, K) tensor of mask quality predictions.
        """
        image_embeddings = inputs["image_embeddings"]
        image_pe = inputs["image_pe"]
        sparse_embeddings = inputs["sparse_embeddings"]
        dense_embeddings = inputs["dense_embeddings"]

        masks, iou_pred = self.predict_masks(
            image_embeddings=image_embeddings,
            image_pe=image_pe,
            sparse_embeddings=sparse_embeddings,
            dense_embeddings=dense_embeddings,
            training=training,
        )

        # Select the correct mask or masks for returning
        if multimask_output:
            return masks[:, 1:], iou_pred[:, 1:]
        else:
            return masks[:, 0:1], iou_pred[:, 0:1]

    def predict_masks(
        self,
        image_embeddings: tf.Tensor,
        image_pe: tf.Tensor,
        sparse_embeddings: tf.Tensor,
        dense_embeddings: tf.Tensor,
        training: bool,
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        """Predicts masks. See `call` for more details."""
        n, h, w, c = tf.unstack(tf.shape(image_embeddings))

        # Concatenate output tokens
        output_tokens = tf.concat([self.iou_token, self.mask_tokens], axis=0)
        output_tokens = tf.expand_dims(output_tokens, axis=0)  # (1, M+1, C)
        output_tokens = tf.tile(output_tokens, (n, 1, 1))  # (N, M+1, C)
        tokens = tf.concat([output_tokens, sparse_embeddings], axis=1)  # (N, M+1+L, C)

        # Run the transformer
        tokens, image_embeddings = self.transformer(
            inputs={
                "point_embeddings": tokens,
                "image_embeddings": image_embeddings + dense_embeddings,
                "image_pe": image_pe,
            },
            training=training,
        )
        iou_token = tokens[:, 0, :]
        mask_tokens = tokens[:, 1 : (1 + self.nb_mask_tokens), :]

        # Upscale mask embeddings and predict masks using the mask tokens
        upscaled_embeddings = self.output_upscaling(image_embeddings, training=training)

        hyper_in_list = [
            self.output_hypernetworks_mlps[j](mask_tokens[:, j, :], training=training)
            for j in range(self.nb_mask_tokens)
        ]
        hyper_in = tf.stack(hyper_in_list, axis=1)

        n, h, w, c = tf.unstack(tf.shape(upscaled_embeddings))
        upscaled_embeddings = tf.reshape(upscaled_embeddings, (n, h * w, c))
        masks = tf.matmul(hyper_in, upscaled_embeddings, transpose_b=True)
        masks = tf.reshape(masks, (n, -1, h, w))

        # Generate mask quality predictions
        iou_pred = self.iou_prediction_head(iou_token)

        return masks, iou_pred


class OutputUpscaling(tf.keras.Model):
    def __init__(
        self,
        embed_dim: int,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.act_layer = act_layer

        norm_layer = norm_layer_factory("layer_norm_eps_1e-6")
        act_layer = act_layer_factory(self.act_layer)

        self.conv1 = tf.keras.layers.Conv2DTranspose(
            filters=self.embed_dim // 4,
            kernel_size=2,
            strides=2,
            use_bias=True,
            name="0",
        )
        self.norm1 = norm_layer(name="1")
        self.act1 = act_layer(name="2")
        self.conv2 = tf.keras.layers.Conv2DTranspose(
            filters=self.embed_dim // 8,
            kernel_size=2,
            strides=2,
            use_bias=True,
            name="3",
        )
        self.act2 = act_layer(name="4")

    def call(self, inputs, training=False):
        x = inputs
        x = self.conv1(x)
        x = self.norm1(x, training=training)
        x = self.act1(x)
        x = self.conv2(x)
        x = self.act2(x)
        return x


class MLP(tf.keras.Model):
    def __init__(
        self,
        hidden_dim: int,
        output_dim: int,
        nb_layers: int,
        sigmoid_output: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.nb_layers = nb_layers
        self.sigmoid_output = sigmoid_output

        filters = [hidden_dim] * (nb_layers - 1) + [output_dim]
        self.blocks = [
            tf.keras.layers.Dense(units=f, use_bias=True, name=f"layers/{j}")
            for j, f in enumerate(filters)
        ]

    def call(self, inputs):
        x = inputs
        for j, block in enumerate(self.blocks):
            x = block(x)
            if j < self.nb_layers - 1:
                x = tf.nn.relu(x)
        if self.sigmoid_output:
            x = tf.nn.sigmoid(x)
        return x



"""
TensorFlow implementation of the Pyramid Vision Transformer

Paper: Pyramid Vision Transformer: A Versatile Backbone for Dense Prediction without
    Convolutions.
Link: https://arxiv.org/pdf/2102.12122.pdf
Official implementation (pytorch): https://github.com/whai362/PVT

Copyright: 2021 Martins Bruveris
"""
from dataclasses import dataclass
from functools import partial
from typing import List, Tuple

import numpy as np
import tensorflow as tf

from tfimm.layers import (
    MLP,
    DropPath,
    PatchEmbeddings,
    act_layer_factory,
    interpolate_pos_embeddings,
    norm_layer_factory,
)
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# model_registry will add each entrypoint fn to this
__all__ = ["PyramidVisionTransformer", "PyramidVisionTransformerConfig"]


@dataclass
class PyramidVisionTransformerConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: Tuple = (4, 2, 2, 2)
    embed_dim: Tuple = (64, 128, 256, 512)
    nb_blocks: Tuple = (3, 4, 6, 3)
    nb_heads: Tuple = (1, 2, 5, 8)
    mlp_ratio: Tuple = (8.0, 8.0, 4.0, 4.0)
    sr_ratio: Tuple = (8, 4, 2, 1)
    qkv_bias: bool = True
    # Regularization
    drop_rate: float = 0.0
    attn_drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Other parameters
    norm_layer: str = "layer_norm_eps_1e-6"
    act_layer: str = "gelu"
    # Parameters for inference
    interpolate_input: bool = False
    crop_pct: float = 0.9
    interpolation: str = "bicubic"
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    first_conv: str = "patch_embed1/proj"
    classifier: str = "head"

    """
    Args:
        nb_classes: Number of classes for classification head
        in_channels: Number of input channels
        input_size: Input image size
        patch_size: Patch size
        embed_dim: Embedding dimension per stage
        nb_blocks: Number of encoder blocks per stage
        nb_heads: Number of self-attention heads per stage
        mlp_ratio: Ratio of mlp hidden dim to embedding dim per stage
        sr_ratio: Spatial reduction ratio
        qkv_bias: Enable bias for qkv if True
        drop_rate: Dropout rate
        attn_drop_rate: Attention dropout rate
        drop_path_rate: Dropout rate for stochastic depth
        norm_layer: Normalization layer
        act_layer: Activation function
    """

    @property
    def nb_tokens(self) -> Tuple:
        """Number of special tokens. Class token is added during the last stage."""
        return 0, 0, 0, 1

    @property
    def grid_size(self) -> Tuple:
        grid_sizes = []
        input_size = self.input_size
        for patch_size in self.patch_size:
            grid_sizes.append(
                (input_size[0] // patch_size, input_size[1] // patch_size)
            )
            input_size = grid_sizes[-1]
        return tuple(grid_sizes)

    @property
    def nb_patches(self) -> Tuple:
        """Number of patches without class token."""
        return tuple(grid_size[0] * grid_size[1] for grid_size in self.grid_size)

    @property
    def transform_weights(self):
        return {
            f"pos_embed{j+1}": partial(
                PyramidVisionTransformer.transform_pos_embed, stage=j
            )
            for j in range(len(self.nb_blocks))
        }


class SpatialReductionAttention(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        sr_ratio: int,
        qkv_bias: bool,
        attn_drop_rate: float,
        proj_drop_rate: float,
        norm_layer: str,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if embed_dim % nb_heads != 0:
            raise ValueError(
                f"embed_dim={embed_dim} should be divisible by nb_heads={nb_heads}."
            )
        self.nb_heads = nb_heads
        self.sr_ratio = sr_ratio
        head_dim = embed_dim // nb_heads
        self.scale = head_dim**-0.5
        self.norm_layer = norm_layer_factory(norm_layer)
        self.act_layer = act_layer_factory(act_layer)

        self.q = tf.keras.layers.Dense(units=embed_dim, use_bias=qkv_bias, name="q")
        self.kv = tf.keras.layers.Dense(
            units=2 * embed_dim, use_bias=qkv_bias, name="kv"
        )
        self.attn_drop = tf.keras.layers.Dropout(rate=attn_drop_rate)
        self.proj = tf.keras.layers.Dense(units=embed_dim, name="proj")
        self.proj_drop = tf.keras.layers.Dropout(rate=proj_drop_rate)

        if sr_ratio > 1:
            self.sr = tf.keras.layers.Conv2D(
                filters=embed_dim,
                kernel_size=sr_ratio,
                strides=sr_ratio,
                name="sr",
            )
            self.norm = self.norm_layer(name="norm")

    def call(self, x, training=False):
        x, grid_size = x  # We need to pass information about spatial shape of input
        batch_size, seq_length, embed_dim = tf.unstack(tf.shape(x))
        head_dim = embed_dim // self.nb_heads
        height, width = tf.unstack(grid_size)

        q = self.q(x)  # (B, N, D)
        q = tf.reshape(q, (batch_size, seq_length, self.nb_heads, -1))
        q = tf.transpose(q, (0, 2, 1, 3))  # (B, H, N, D/H)

        # Spatial reduction happens here. In the last stage, after adding the class
        # token, `sr_ratio=1`. Note that after adding the class token, the resize
        # operation will fail.
        if self.sr_ratio > 1:
            x = tf.reshape(x, (batch_size, height, width, embed_dim))  # (B, h, w, D)
            x = self.sr(x)  # (B, h', w', D)
            x = tf.reshape(x, (batch_size, -1, embed_dim))  # (B, N', D)
            x = self.norm(x, training=training)

        kv = self.kv(x)  # (B, N', 2 * D)
        kv = tf.reshape(kv, (batch_size, -1, 2, self.nb_heads, head_dim))
        kv = tf.transpose(kv, (2, 0, 3, 1, 4))  # (2, B, H, N', D/H)
        k, v = kv[0], kv[1]

        attn = self.scale * tf.linalg.matmul(q, k, transpose_b=True)  # (B, H, N, N')
        attn = tf.nn.softmax(attn, axis=-1)  # (B, H, N, N')
        attn = self.attn_drop(attn, training=training)

        x = tf.linalg.matmul(attn, v)  # (B, H, N, D/H)
        x = tf.transpose(x, (0, 2, 1, 3))  # (B, N, H, D/H)
        x = tf.reshape(x, (batch_size, seq_length, -1))  # (B, N, D)

        x = self.proj(x)
        x = self.proj_drop(x, training=training)

        return x


class Block(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        mlp_ratio: float,
        sr_ratio: int,
        qkv_bias: bool,
        drop_rate: float,
        attn_drop_rate: float,
        drop_path_rate: float,
        norm_layer: str,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.norm_layer = norm_layer_factory(norm_layer)

        self.norm1 = self.norm_layer(name="norm1")
        self.attn = SpatialReductionAttention(
            embed_dim=embed_dim,
            nb_heads=nb_heads,
            sr_ratio=sr_ratio,
            qkv_bias=qkv_bias,
            proj_drop_rate=drop_rate,
            attn_drop_rate=attn_drop_rate,
            # We don't pass the norm_layer parameter, because that is set to
            # layer_norm_eps_1e-6 in this module
            norm_layer="layer_norm",
            act_layer=act_layer,
            name="attn",
        )
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.norm2 = self.norm_layer(name="norm2")
        self.mlp = MLP(
            hidden_dim=int(embed_dim * mlp_ratio),
            embed_dim=embed_dim,
            drop_rate=drop_rate,
            act_layer=act_layer,
            name="mlp",
        )

    def call(self, x, training=False):
        x, grid_size = x

        shortcut = x
        x = self.norm1(x, training=training)
        x = self.attn((x, grid_size), training=training)
        x = self.drop_path(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp(x, training=training)
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return x


@keras_serializable
class PyramidVisionTransformer(tf.keras.Model):
    cfg_class = PyramidVisionTransformerConfig

    def __init__(self, cfg: PyramidVisionTransformerConfig, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(**kwargs)
        self.cfg = cfg

        self.patch_embed = None
        self.pos_embed = None
        self.pos_drop = None
        self.blocks = None
        self.norm = None
        self.cls_token = None
        self.head = None

    def build(self, input_shape):
        norm_layer = norm_layer_factory(self.cfg.norm_layer)

        # Stochastic depth
        dpr = np.linspace(0, self.cfg.drop_path_rate, sum(self.cfg.nb_blocks))
        block_idx = 0

        self.patch_embed = []
        self.pos_embed = []
        self.pos_drop = []
        self.blocks = []
        nb_stages = len(self.cfg.nb_blocks)
        for j in range(nb_stages):
            self.patch_embed.append(
                PatchEmbeddings(
                    patch_size=self.cfg.patch_size[j],
                    embed_dim=self.cfg.embed_dim[j],
                    norm_layer="layer_norm",
                    name=f"patch_embed{j+1}",
                )
            )
            nb_tokens = self.cfg.nb_patches[j] + self.cfg.nb_tokens[j]
            self.pos_embed.append(
                self.add_weight(
                    shape=(1, nb_tokens, self.cfg.embed_dim[j]),
                    initializer="zeros",
                    trainable=True,
                    name=f"pos_embed{j+1}",
                )
            )
            self.pos_drop.append(tf.keras.layers.Dropout(rate=self.cfg.drop_rate))
            for k in range(self.cfg.nb_blocks[j]):
                self.blocks.append(
                    Block(
                        embed_dim=self.cfg.embed_dim[j],
                        nb_heads=self.cfg.nb_heads[j],
                        mlp_ratio=self.cfg.mlp_ratio[j],
                        sr_ratio=self.cfg.sr_ratio[j],
                        qkv_bias=self.cfg.qkv_bias,
                        drop_rate=self.cfg.drop_rate,
                        attn_drop_rate=self.cfg.attn_drop_rate,
                        drop_path_rate=dpr[block_idx],
                        norm_layer=self.cfg.norm_layer,
                        act_layer=self.cfg.act_layer,
                        name=f"block{j+1}/{k}",
                    )
                )
                block_idx += 1
        self.norm = norm_layer(name="norm")
        self.cls_token = self.add_weight(
            shape=(1, 1, self.cfg.embed_dim[-1]),
            initializer="zeros",
            trainable=True,
            name="cls_token",
        )
        self.head = (
            tf.keras.layers.Dense(units=self.cfg.nb_classes, name="head")
            if self.cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        names = []
        k = 0
        for j in range(len(self.cfg.nb_blocks)):
            names += [f"patch_embedding_{j}", f"pos_embedding_{j}"]
            for _ in range(self.cfg.nb_blocks[j]):
                names.append(f"block_{k}")
                k += 1
            names.append(f"stage_{j}")
        names += ["features_all", "features", "logits"]
        return names

    # @torch.jit.ignore
    # def no_weight_decay(self):
    #     # return {'pos_embed', 'cls_token'} # has pos_embed may be better
    #     return {'cls_token'}

    def transform_pos_embed(
        self, src_weights, target_cfg: PyramidVisionTransformerConfig, stage: int
    ):
        return interpolate_pos_embeddings(
            pos_embed=self.pos_embed[stage],
            src_grid_size=self.cfg.grid_size[stage],
            tgt_grid_size=target_cfg.grid_size[stage],
            nb_tokens=self.cfg.nb_tokens[stage],
        )

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        batch_size = tf.shape(x)[0]

        nb_stages = len(self.cfg.nb_blocks)
        k = 0
        for j in range(nb_stages):
            x, grid_size = self.patch_embed[j](x, training=training, return_shape=True)
            features[f"patch_embedding_{j}"] = x

            if j == nb_stages - 1:
                cls_token = tf.repeat(self.cls_token, repeats=batch_size, axis=0)
                x = tf.concat((cls_token, x), axis=1)
            if not self.cfg.interpolate_input:
                x = x + self.pos_embed[j]
            else:
                pos_embed = interpolate_pos_embeddings(
                    self.pos_embed[j],
                    src_grid_size=self.cfg.grid_size[j],
                    tgt_grid_size=grid_size,
                    nb_tokens=self.cfg.nb_tokens[j],
                )
                x = x + pos_embed
            x = self.pos_drop[j](x, training=training)
            features[f"pos_embedding_{j}"] = x

            for _ in range(self.cfg.nb_blocks[j]):
                x = self.blocks[k]((x, grid_size), training=training)
                features[f"block_{k}"] = x
                k += 1

            if j != nb_stages - 1:
                # Reshape to image form, ready for patch embedding
                x = tf.reshape(x, (batch_size, *grid_size, -1))
            features[f"stage_{j}"] = x

        x = self.norm(x, training=training)
        features["features_all"] = x
        x = x[:, 0]
        features["features"] = x
        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x, training=training)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def pvt_tiny():
    cfg = PyramidVisionTransformerConfig(
        name="pvt_tiny",
        url="[pytorch]https://github.com/whai362/PVT/releases/download/v2/pvt_tiny.pth",
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(2, 2, 2, 2),
    )
    return PyramidVisionTransformer, cfg


@register_model
def pvt_small():
    cfg = PyramidVisionTransformerConfig(
        name="pvt_small",
        url="[pytorch]https://github.com/whai362/PVT/releases/download/v2/pvt_small.pth",  # noqa: E501
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(3, 4, 6, 3),
    )
    return PyramidVisionTransformer, cfg


@register_model
def pvt_medium():
    cfg = PyramidVisionTransformerConfig(
        name="pvt_medium",
        url="[pytorch]https://github.com/whai362/PVT/releases/download/v2/pvt_medium.pth",  # noqa: E501
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(3, 4, 18, 3),
    )
    return PyramidVisionTransformer, cfg


@register_model
def pvt_large():
    cfg = PyramidVisionTransformerConfig(
        name="pvt_large",
        url="[pytorch]https://github.com/whai362/PVT/releases/download/v2/pvt_large.pth",  # noqa: E501
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(3, 8, 27, 3),
    )
    return PyramidVisionTransformer, cfg



_classes = {}
_cfg_classes = {}


def cfg_serializable(cls):
    """
    Registers the configuration datatype. The configuration class can either be
    associated to a parent class (whose configuration it is) or be standalone.

    In the first case we decorate a class with a `cfg_class` attribute
    ```
    @dataclass
    class DatasetConfig:
        path: str

    @cfg_serializable
    class Dataset:
        cfg_class = DatasetConfig

        def __init__(self, cfg: DatasetConfig):
            ...
    ```

    In the second case, we have only a config class (see `Timekeeping`), decorated
    directly.
    ```
    @cfg_serializable
    @dataclass
    class Timekeeping
        nb_epochs: int
    ```

    We will use the information to parse configurations from file or from command line
    arguments.
    """
    cls_name = cls.__name__

    if hasattr(cls, "cfg_class"):
        # Here we are dealing with a class, that has an associated config class.
        # First we associate the class name with the class and the config type
        _classes[cls_name] = cls
        _cfg_classes[cls_name] = cls.cfg_class

        # We also register the config class itself
        cfg_class_name = cls.cfg_class.__name__
        _cfg_classes[cfg_class_name] = cls.cfg_class
    else:
        # Here we are dealing with a standalone configuration class
        _cfg_classes[cls_name] = cls
    return cls


def get_cfg_class(cls):
    """Retrieves the configuration datatype associated to a class."""
    return _cfg_classes[cls]


def get_class(cls):
    """Retrieves the registered class."""
    return _classes[cls]



import tensorflow as tf

from tfimm.layers import norm_layer_factory

from .common import MLPBlock


class TwoWayTransformer(tf.keras.Model):
    def __init__(
        self,
        embed_dim: int,
        nb_blocks: int,
        nb_heads: int,
        mlp_dim: int,
        attention_downsample_rate: int,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.nb_blocks = nb_blocks
        self.nb_heads = nb_heads
        self.mlp_dim = mlp_dim
        self.attention_downsample_rate = attention_downsample_rate
        self.act_layer = act_layer

        norm_layer = norm_layer_factory("layer_norm")

        self.blocks = [
            TwoWayAttentionBlock(
                embed_dim=self.embed_dim,
                nb_heads=self.nb_heads,
                mlp_dim=self.mlp_dim,
                attention_downsample_rate=self.attention_downsample_rate,
                skip_first_layer_pe=j == 0,
                act_layer=self.act_layer,
                name=f"layers/{j}",
            )
            for j in range(self.nb_blocks)
        ]
        self.final_attn_token_to_image = Attention(
            embed_dim=self.embed_dim,
            nb_heads=self.nb_heads,
            downsample_rate=self.attention_downsample_rate,
            name="final_attn_token_to_image",
        )
        self.norm_final_attn = norm_layer(name="norm_final_attn")

    def call(self, inputs, training=False):
        """
        Args:
            inputs: Dictionary with the following entries.
                point_embeddings: Point embedding, should have shape (B, N, embed_dim)
                        for any N.
                image_embeddings: Image to attend to. Should have the shape
                    (B, H, W, embed_dim) for any (H, W).
                image_pe: Positional encoding to add to the image. Must have the same
                    shape as image_embedding.
            training: Training or inference mode?

        Returns:
            The processed point_embedding, same shape as input.
            The processed image_embedding, same shape as input. Note that this differs
                from PyTorch, where the output has shape (B, H*W, embed_dim).
        """
        image_embeddings = inputs["image_embeddings"]
        image_pe = inputs["image_pe"]
        point_embeddings = inputs["point_embeddings"]
        b, h, w, c = tf.unstack(tf.shape(image_embeddings))

        image_embedding = tf.reshape(image_embeddings, (b, -1, c))  # (B, H*W, C)
        image_pe = tf.reshape(image_pe, (b, -1, c))  # (B, H*W, C)

        # Prepare queries
        queries = point_embeddings
        keys = image_embedding

        # Apply transformer blocks
        for block in self.blocks:
            queries, keys = block(
                {"q": queries, "k": keys, "q_pe": point_embeddings, "k_pe": image_pe},
                training=training,
            )

        # Apply final attention layer from the points to the image
        attn = self.final_attn_token_to_image(
            {"q": queries + point_embeddings, "k": keys + image_pe, "v": keys},
            training=training,
        )
        queries = queries + attn
        queries = self.norm_final_attn(queries, training=training)

        # Reshape back to (B, H, W, C)
        keys = tf.reshape(keys, (b, h, w, c))

        return queries, keys


class TwoWayAttentionBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        mlp_dim: int,
        attention_downsample_rate: int,
        skip_first_layer_pe: bool,
        act_layer: str,
        **kwargs,
    ):
        """
        A transformer block with four layers:
            (1) self-attention of sparse inputs,
            (2) cross attention of sparse inputs to dense inputs,
            (3) mlp block on sparse inputs, and
            (4) cross attention of dense inputs to sparse inputs.

        Args:
          embed_dim: the channel dimension of the embeddings
          num_heads: the number of heads in the attention layers
          mlp_dim: the hidden dimension of the mlp block
          skip_first_layer_pe: skip the PE on the first layer
          act_layer: the activation of the mlp block
        """
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads
        self.mlp_dim = mlp_dim
        self.attention_downsample_rate = attention_downsample_rate
        self.skip_first_layer_pe = skip_first_layer_pe
        self.act_layer = act_layer

        norm_layer = norm_layer_factory("layer_norm")

        self.self_attn = Attention(
            embed_dim=embed_dim, nb_heads=nb_heads, downsample_rate=1, name="self_attn"
        )
        self.norm1 = norm_layer(name="norm1")

        self.cross_attn_token_to_image = Attention(
            embed_dim=embed_dim,
            nb_heads=nb_heads,
            downsample_rate=attention_downsample_rate,
            name="cross_attn_token_to_image",
        )
        self.norm2 = norm_layer(name="norm2")

        self.mlp = MLPBlock(
            hidden_dim=self.mlp_dim,
            embed_dim=self.embed_dim,
            act_layer=self.act_layer,
            drop_rate=0.0,
            name="mlp",
        )
        self.norm3 = norm_layer(name="norm3")

        self.cross_attn_image_to_token = Attention(
            embed_dim=embed_dim,
            nb_heads=nb_heads,
            downsample_rate=attention_downsample_rate,
            name="cross_attn_image_to_token",
        )
        self.norm4 = norm_layer(name="norm4")

    def call(self, inputs, training=False):
        q, k, q_pe, k_pe = inputs["q"], inputs["k"], inputs["q_pe"], inputs["k_pe"]

        # Self-attention block
        if self.skip_first_layer_pe:
            q = self.self_attn({"q": q, "k": q, "v": q}, training=training)
        else:
            attn = self.self_attn({"q": q + q_pe, "k": q + q_pe, "v": q})
            q = q + attn
        q = self.norm1(q, training=training)

        # Cross-attention block, tokens attending to image embedding
        attn = self.cross_attn_token_to_image(
            {"q": q + q_pe, "k": k + k_pe, "v": k}, training=training
        )
        q = q + attn
        q = self.norm2(q, training=training)

        # MLP block
        mlp = self.mlp(q, training=training)
        q = q + mlp
        q = self.norm3(q, training=training)

        # Cross-attention block, image embeddings attending to tokens
        attn = self.cross_attn_image_to_token(
            {"q": k + k_pe, "k": q + q_pe, "v": q}, training=training
        )
        k = k + attn
        k = self.norm4(k, training=training)

        return q, k


class Attention(tf.keras.layers.Layer):
    """
    An attention layer that allows for downscaling the size of the embedding after
    projection to queries, keys, and values.
    """

    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        downsample_rate: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads
        self.downsample_rate = downsample_rate

        internal_dim = self.embed_dim // self.downsample_rate
        self.q_proj = tf.keras.layers.Dense(internal_dim, use_bias=True, name="q_proj")
        self.k_proj = tf.keras.layers.Dense(internal_dim, use_bias=True, name="k_proj")
        self.v_proj = tf.keras.layers.Dense(internal_dim, use_bias=True, name="v_proj")
        self.out_proj = tf.keras.layers.Dense(
            units=self.embed_dim, use_bias=True, name="out_proj"
        )

    def _separate_heads(self, x: tf.Tensor):
        b, m, c = tf.shape(x)  # (B, M, C)
        x = tf.reshape(x, (b, m, self.nb_heads, c // self.nb_heads))  # (B, M, Hd, C/Hd)
        x = tf.transpose(x, (0, 2, 1, 3))  # (B, Hd, M, C/Hd)
        return x

    def _recombine_heads(self, x: tf.Tensor):
        # Shape of x is (B, Hd, M, C/Hd)
        batch_size, _, seq_length, _ = tf.unstack(tf.shape(x))
        x = tf.transpose(x, (0, 2, 1, 3))  # (B, M, Hd, C/Hd)
        x = tf.reshape(x, (batch_size, seq_length, -1))  # (B, M, C)
        return x

    def call(self, inputs):
        q, k, v = inputs["q"], inputs["k"], inputs["v"]

        # Input projections
        q = self.q_proj(q)
        k = self.k_proj(k)
        v = self.v_proj(v)

        # Separate into heads
        q = self._separate_heads(q)
        k = self._separate_heads(k)
        v = self._separate_heads(v)

        # Attention
        d = tf.shape(q)[-1]  # D=C/Hd
        attn = tf.matmul(q, k, transpose_b=True)  # (B, Hd, M, M)
        attn = attn / tf.sqrt(tf.cast(d, tf.float32))
        attn = tf.nn.softmax(attn, axis=-1)  # (B, Hd, M, M)

        # Get output
        x = tf.matmul(attn, v)  # (B, Hd, M, C/Hd)
        x = self._recombine_heads(x)  # (B, M, C)
        x = self.out_proj(x)

        return x



import logging
import os
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Union

try:
    import wandb
except ImportError:
    wandb = None
    logging.info("Could not import `wandb`. Logging to W&B not possible.")

from . import config
from .registry import get_class
from .utils import setup_logging


@dataclass
class ExperimentConfig:
    trainer: Any
    trainer_class: str
    timekeeping: Any
    timekeeping_class: str
    problem: Any
    problem_class: str
    # Data
    train_dataset: Any
    train_dataset_class: str
    val_dataset: Any = None
    val_dataset_class: str = ""
    # One of 0, 10, 20, 30, 40, 50: NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
    logging_level: int = 30
    # W&B parameters
    log_wandb: bool = False
    experiment_name: str = "default"  # Experiment name in W&B
    project_name: str = "default"  # Project name in W&B
    entity: str = "default"  # Entity in W&B
    # Configuration file
    cfg_file: str = ""


def run(cfg: Union[ExperimentConfig, dict], parse_args: bool = True):
    """
    Runs experiment defined by `cfg`.

    If `parse_args=True`, we will parse command line arguments. We will always parse
    the cfg_file.
    """
    # If we want to parse command line args, we provide None to `parse_args`, which
    # will default to using argv[1:]. By providing the empty list it will not look
    # for other arguments to parse.
    args = None if parse_args else []
    cfg = config.parse_args(cfg, cfg_class=ExperimentConfig, args=args)

    # Configure logging level
    setup_logging(cfg.logging_level)

    # Print config to stdout
    logging.info("Running with arguments")
    config.pprint(cfg)

    # Save config to file
    ckpt_dir = getattr(cfg.trainer, "ckpt_dir", "")
    if ckpt_dir:
        config.dump_config(cfg, Path(ckpt_dir) / "config.yaml")

    # Prepare W&B
    if cfg.log_wandb:
        wandb.init(
            dir=tempfile.gettempdir(),
            config=config.deep_to_flat(config.to_dict_format(cfg)),
            entity=cfg.entity,
            project=cfg.project_name,
            name=cfg.experiment_name,
            resume=False,
        )
        if wandb.run.sweep_id:
            # If this run is part of a W&B hyperparameter sweep, we need to add
            # suffixes to the run names and checkpoint directories, because otherwise
            # all runs in the sweep will have the same name and the checkpoints will
            # overwrite each other.
            logging.info(f"Job running as part of sweep {wandb.run.sweep_id}.")
            ckpt_dir = getattr(cfg.trainer, "ckpt_dir", "")
            if ckpt_dir:
                setattr(cfg.trainer, "ckpt_dir", os.path.join(ckpt_dir, wandb.run.id))
            wandb.run.name = wandb.run.name + f"{wandb.run.id}"
            wandb.run.save()

    # Construct constituent objects
    train_ds = get_class(cfg.train_dataset_class)(cfg=cfg.train_dataset)
    val_ds = (
        get_class(cfg.val_dataset_class)(cfg=cfg.val_dataset)
        if cfg.val_dataset_class
        else None
    )
    problem = get_class(cfg.problem_class)(cfg=cfg.problem, timekeeping=cfg.timekeeping)
    trainer = get_class(cfg.trainer_class)(
        problem=problem,
        train_ds=train_ds,
        val_ds=val_ds,
        timekeeping=cfg.timekeeping,
        log_wandb=cfg.log_wandb,
        cfg=cfg.trainer,
    )

    # Do the training!
    trainer.train()



"""
We provide an implementation and pretrained weights for the EfficientNet family of
models.

Paper: EfficientNet: Rethinking Model Scaling for CNNs.
`[arXiv:1905.11946] <https://arxiv.org/abs/1905.11946>`_.

This code and weights have been ported from the
`timm <https://github.com/rwightman/pytorch-image-models>`_ implementation. It does mean
that some model weights have undergone the journey from TF (original weights from the
Google Brain team) to PyTorch (timm library) back to TF (tfimm port of timm).

The following models are available.

* MobileNet-V2 models. These models correspond to the ``mobilenetv2_...`` models in
  timm.

  * ``mobilenet_v2_{050, 100, 140}``. These are MobileNet-V2 models with channel
    multiplier set to 0.5, 1.0 and 1.4 respectively.
  * ``mobilenet_v2_{110d, 120d}``. These are MobileNet-V2 models with (channel, depth)
    multipliers set to (1.1, 1.2) and (1.2, 1.4) respectively.

* Original EfficientNet models. These models correspond to the models ``tf_...`` in
  timm.

  * ``efficientnet_{b0, b1, b2, b3, b4, b5, b6, b7, b8}``

* EfficientNet AdvProp models, trained with adversarial examples. These models
  correspond to the ``tf_...`` models in timm.

  * ``efficientnet_{b0, ..., b8}_ap``

* EfficientNet NoisyStudent models, trained via semi-supervised learning. These models
  correspond to the ``tf_...`` models in timm.

  * ``efficientnet_{b0, ..., b7}_ns``
  * ``efficientnet_l2_ns_475``
  * ``efficientnet_l2``

* PyTorch versions of the EfficientNet models. These models use symmetric padding rather
  than "same" padding that is default in TF. They correspond to the ``efficientnet_...``
  models in timm.

  * ``pt_efficientnet_{b0, ..., b4}``

* EfficientNet-EdgeTPU models, optimized for inference on Google's Edge TPU hardware.
  These models correspond to the ``tf_...`` models in timm.

  * ``efficientnet_es``
  * ``efficientnet_em``
  * ``efficientnet_el``

* EfficientNet-Lite models, optimized for inference on mobile devices, CPUs and GPUs.
  These models correspond to the ``tf_...`` models in timm.

  * ``efficientnet_lite0``
  * ``efficientnet_lite1``
  * ``efficientnet_lite2``
  * ``efficientnet_lite3``
  * ``efficientnet_lite4``

* EfficientNet-V2 models. These models correspond to the ``tf_...`` models in timm.

  * ``efficientnet_v2_b0``
  * ``efficientnet_v2_b1``
  * ``efficientnet_v2_b2``
  * ``efficientnet_v2_b3``
  * ``efficientnet_v2_s``
  * ``efficientnet_v2_m``
  * ``efficientnet_v2_l``

* EfficientNet-V2 models, pretrained on ImageNet-21k, fine-tuned on ImageNet-1k. These
  models correspond to the ``tf_...`` models in timm.

  * ``efficientnet_v2_s_in21ft1k``
  * ``efficientnet_v2_m_in21ft1k``
  * ``efficientnet_v2_l_in21ft1k``
  * ``efficientnet_v2_xl_in21ft1k``

* EfficientNet-V2 models, pretrained on ImageNet-21k. These models correspond to the
  ``tf_...`` models in timm.

  * ``efficientnet_v2_s_in21k``
  * ``efficientnet_v2_m_in21k``
  * ``efficientnet_v2_l_in21k``
  * ``efficientnet_v2_xl_in21k``
"""
# Hacked together by / Copyright 2019, Ross Wightman
# Copyright 2022 Marting Bruveris
from collections import OrderedDict
from dataclasses import dataclass
from functools import partial
from typing import List, Tuple

import tensorflow as tf

from tfimm.architectures.efficientnet_builder import (
    EfficientNetBuilder,
    decode_architecture,
    round_channels,
)
from tfimm.layers import act_layer_factory, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import (
    IMAGENET_DEFAULT_MEAN,
    IMAGENET_DEFAULT_STD,
    IMAGENET_INCEPTION_MEAN,
    IMAGENET_INCEPTION_STD,
)

from .efficientnet_blocks import create_conv2d

# Model registry will add each entrypoint fn to this
__all__ = ["EfficientNetConfig", "EfficientNet"]

# TODO: Fix list_timm_models with two different model names.


@dataclass
class EfficientNetConfig(ModelConfig):
    """
    Configuration class for EfficientNet models.

    Parameters:
        name: Name of the model.
        url: URL for pretrained weights.
        nb_classes: Number of classes for classification head.
        in_channels: Number of input image channels.
        input_size: Input image size (height, width)

        stem_size: Number of filters in first convolution.
        architecture: Tuple of tuple of strings defining the architecture of residual
            blocks. The outer tuple defines the stages while the inner tuple defines
            the blocks per stage.
        channel_multiplier: Multiplier for channel scaling. One of the three dimensions
            of EfficientNet scaling.
        depth_multiplier: Multiplier for depth scaling. One of the three dimensions of
            EfficientNet scaling.
        fix_first_last:  Fix first and last block depths when multiplier is applied.
        nb_features: Number of features before the classifier layer.

        drop_rate: Dropout rate.
        drop_path_rate: Dropout rate for stochastic depth.

        norm_layer: Normalization layer. See :func:`~norm_layer_factory` for possible
            values.
        act_layer: Activation function. See :func:`~act_layer_factory` for possible
            values.
        padding: Type of padding to use for convolutional layers. Can be one of
            "same", "valid" or "symmetric" (PyTorch-style symmetric padding).

        crop_pct: Crop percentage for ImageNet evaluation.
        interpolation: Interpolation method for ImageNet evaluation.
        mean: Defines preprocessing function. If ``x`` is an image with pixel values
            in (0, 1), the preprocessing function is ``(x - mean) / std``.
        std: Defines preprpocessing function.

        first_conv: Name of first convolutional layer. Used by
            :func:`~tfimm.create_model` to adapt the number in input channels when
            loading pretrained weights.
        classifier: Name of classifier layer. Used by :func:`~tfimm.create_model` to
            adapt the classifier when loading pretrained weights.
    """

    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    # Architecture
    stem_size: int = 32
    architecture: Tuple[Tuple[str, ...], ...] = ()
    channel_multiplier: float = 1.0
    depth_multiplier: float = 1.0
    fix_first_last: bool = False
    nb_features: int = 1280
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Other params
    norm_layer: str = "batch_norm"
    act_layer: str = "swish"
    padding: str = "symmetric"  # One of "symmetric", "same", "valid"
    # Parameters for inference
    crop_pct: float = 0.875
    interpolation: str = "bicubic"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "conv_stem"
    classifier: str = "classifier"


@keras_serializable
class EfficientNet(tf.keras.Model):
    """
    Generic EfficientNet implementation supporting depth and width scaling and flexible
    architecture definitions, including

    * EfficientNet B0-B7.

    Parameters:
        cfg: Configuration class for the model.
        **kwargs: Arguments are passed to ``tf.keras.Model``.
    """

    cfg_class = EfficientNetConfig

    def __init__(self, cfg: EfficientNetConfig, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(**kwargs)
        self.cfg = cfg
        self.norm_layer = norm_layer_factory(cfg.norm_layer)
        self.act_layer = act_layer_factory(cfg.act_layer)

        # Stem
        self.conv_stem = create_conv2d(
            filters=cfg.stem_size,
            kernel_size=3,
            strides=2,
            padding=cfg.padding,
            name="conv_stem",
        )
        self.bn1 = self.norm_layer(name="bn1")
        self.act1 = self.act_layer()

        # Middle stages (IR/ER/DS Blocks)
        builder = EfficientNetBuilder(
            output_stride=32,
            channel_multiplier=cfg.channel_multiplier,
            padding=cfg.padding,
            act_layer=cfg.act_layer,
            norm_layer=cfg.norm_layer,
            drop_path_rate=cfg.drop_path_rate,
        )
        architecture = decode_architecture(
            architecture=cfg.architecture,
            depth_multiplier=cfg.depth_multiplier,
            depth_truncation="ceil",
            experts_multiplier=1,
            fix_first_last=cfg.fix_first_last,
            group_size=None,
        )
        self.blocks = builder(architecture)

        # Head
        self.conv_head = create_conv2d(
            filters=cfg.nb_features,
            kernel_size=1,
            padding=cfg.padding,
            name="conv_head",
        )
        self.bn2 = self.norm_layer(name="bn2")
        self.act2 = self.act_layer()

        # Pooling + Classifier
        self.pool = tf.keras.layers.GlobalAveragePooling2D()
        self.flatten = tf.keras.layers.Flatten()
        self.drop = tf.keras.layers.Dropout(rate=cfg.drop_rate)
        self.classifier = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="classifier")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        """Returns a tensor of the correct shape for inference."""
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        """
        Names of features, returned when calling ``call`` with ``return_features=True``.
        """
        _, features = self(self.dummy_inputs, return_features=True)
        return list(features.keys())

    def forward_features(
        self, x, training: bool = False, return_features: bool = False
    ):
        """
        Forward pass through model, excluding the classifier layer. This function is
        useful if the model is used as input for downstream tasks such as object
        detection.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.conv_stem(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        features["stem"] = x

        for key, block in self.blocks.items():
            x = block(x, training=training)
            features[key] = x

        x = self.conv_head(x)
        x = self.bn2(x, training=training)
        x = self.act2(x)
        features["conv_features"] = x

        return (x, features) if return_features else x

    def call(self, x, training: bool = False, return_features: bool = False):
        """
        Forward pass through the full model.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x

        x = self.pool(x)
        x = self.flatten(x)
        features["features"] = x

        x = self.drop(x, training=training)
        x = self.classifier(x)
        features["logits"] = x
        return (x, features) if return_features else x


def _mobilenet_v2_cfg(
    name: str,
    timm_name: str,
    channel_multiplier: float = 1.0,
    depth_multiplier: float = 1.0,
    fix_stem_head: bool = False,
    crop_pct: float = 0.875,
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN,
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD,
):
    """
    Creates the config for a MobileNet-v2 model.

    Ref impl: https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet_v2.py  # noqa: E501
    Paper: https://arxiv.org/abs/1801.04381

    Args:
        name: Model name
        timm_name: Name of model in TIMM
        channel_multiplier: Multiplier for channel scaling.
        depth_multiplier: Multiplier for depth scaling.
        fix_stem_head: Scale stem channels and number of features or not
        crop_pct: Crop percentage for ImageNet evaluation
        mean: Defines preprocessing function.
        std: Defines preprpocessing function.
    """
    round_channels_fn = partial(round_channels, multiplier=channel_multiplier)
    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        stem_size=32 if fix_stem_head else round_channels_fn(32),
        architecture=(
            ("ds_r1_k3_s1_c16",),
            ("ir_r2_k3_s2_e6_c24",),
            ("ir_r3_k3_s2_e6_c32",),
            ("ir_r4_k3_s2_e6_c64",),
            ("ir_r3_k3_s1_e6_c96",),
            ("ir_r3_k3_s2_e6_c160",),
            ("ir_r1_k3_s1_e6_c320",),
        ),
        channel_multiplier=channel_multiplier,
        depth_multiplier=depth_multiplier,
        fix_first_last=fix_stem_head,
        nb_features=1280 if fix_stem_head else max(1280, round_channels_fn(1280)),
        norm_layer="batch_norm",
        act_layer="relu6",
        crop_pct=crop_pct,
        mean=mean,
        std=std,
    )
    return cfg


@register_model
def mobilenet_v2_050():
    """MobileNet-V2 with 0.50 channel multiplier"""
    cfg = _mobilenet_v2_cfg(
        name="mobilenet_v2_050",
        timm_name="mobilenetv2_050",
        channel_multiplier=0.50,
    )
    return EfficientNet, cfg


@register_model
def mobilenet_v2_100():
    """MobileNet-V2 with 1.0 channel multiplier"""
    cfg = _mobilenet_v2_cfg(
        name="mobilenet_v2_100",
        timm_name="mobilenetv2_100",
        channel_multiplier=1.0,
    )
    return EfficientNet, cfg


@register_model
def mobilenet_v2_140():
    """MobileNet-V2 with 1.4 channel multiplier"""
    cfg = _mobilenet_v2_cfg(
        name="mobilenet_v2_140",
        timm_name="mobilenetv2_140",
        channel_multiplier=1.4,
    )
    return EfficientNet, cfg


@register_model
def mobilenet_v2_110d():
    """MobileNet-V2 with 1.1 channel and 1.2 depth multiplier"""
    cfg = _mobilenet_v2_cfg(
        name="mobilenet_v2_110d",
        timm_name="mobilenetv2_110d",
        channel_multiplier=1.1,
        depth_multiplier=1.2,
        fix_stem_head=True,
    )
    return EfficientNet, cfg


@register_model
def mobilenet_v2_120d():
    """MobileNet-V2 with 1.2 channel and 1.4 depth multiplier"""
    cfg = _mobilenet_v2_cfg(
        name="mobilenet_v2_120d",
        timm_name="mobilenetv2_120d",
        channel_multiplier=1.2,
        depth_multiplier=1.4,
        fix_stem_head=True,
    )
    return EfficientNet, cfg


def _efficientnet_cfg(
    name: str,
    timm_name: str,
    variant: str,
    input_size: Tuple[int, int],
    framework: str,
    crop_pct: float,
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN,
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD,
):
    """
    Creates the config for an EfficientNet model.

    Ref impl: https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/efficientnet_model.py  # noqa: E501
    Paper: https://arxiv.org/abs/1905.11946

    Args:
        name: Model name
        timm_name: Model name in TIMM
        variant: Model variant, e.g., "b0", etc.
        framework: "tf" or "pytorch" for BN params and padding
        crop_pct: Crop percentage for ImageNet evaluation
    """
    assert framework in {"tf", "pytorch"}

    param_dict = {
        "b0": (1.0, 1.0, 0.2),
        "b1": (1.0, 1.1, 0.2),
        "b2": (1.1, 1.2, 0.3),
        "b3": (1.2, 1.4, 0.3),
        "b4": (1.4, 1.8, 0.4),
        "b5": (1.6, 2.2, 0.4),
        "b6": (1.8, 2.6, 0.5),
        "b7": (2.0, 3.1, 0.5),
        "b8": (2.2, 3.6, 0.5),
        "l2": (4.3, 5.3, 0.5),
    }
    channel_multiplier, depth_multiplier, drop_rate = param_dict[variant]

    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        input_size=input_size,
        stem_size=round_channels(32, multiplier=channel_multiplier),
        architecture=(
            ("ds_r1_k3_s1_e1_c16_se0.25",),
            ("ir_r2_k3_s2_e6_c24_se0.25",),
            ("ir_r2_k5_s2_e6_c40_se0.25",),
            ("ir_r3_k3_s2_e6_c80_se0.25",),
            ("ir_r3_k5_s1_e6_c112_se0.25",),
            ("ir_r4_k5_s2_e6_c192_se0.25",),
            ("ir_r1_k3_s1_e6_c320_se0.25",),
        ),
        channel_multiplier=channel_multiplier,
        depth_multiplier=depth_multiplier,
        nb_features=round_channels(1280, multiplier=channel_multiplier),
        drop_rate=drop_rate,
        drop_path_rate=drop_rate,
        norm_layer="batch_norm_tf" if framework == "tf" else "batch_norm",
        act_layer="swish",
        padding="same" if framework == "tf" else "symmetric",
        crop_pct=crop_pct,
        mean=mean,
        std=std,
    )
    return cfg


@register_model
def efficientnet_b0():
    """EfficientNet-B0. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b0",
        timm_name="tf_efficientnet_b0",
        variant="b0",
        input_size=(224, 224),
        framework="tf",
        crop_pct=0.875,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b1():
    """EfficientNet-B1. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b1",
        timm_name="tf_efficientnet_b1",
        variant="b1",
        input_size=(240, 240),
        framework="tf",
        crop_pct=0.882,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b2():
    """EfficientNet-B2. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b2",
        timm_name="tf_efficientnet_b2",
        variant="b2",
        input_size=(260, 260),
        framework="tf",
        crop_pct=0.890,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b3():
    """EfficientNet-B3. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b3",
        timm_name="tf_efficientnet_b3",
        variant="b3",
        input_size=(300, 300),
        framework="tf",
        crop_pct=0.904,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b4():
    """EfficientNet-B4. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b4",
        timm_name="tf_efficientnet_b4",
        variant="b4",
        input_size=(380, 380),
        framework="tf",
        crop_pct=0.922,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b5():
    """EfficientNet-B5. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b5",
        timm_name="tf_efficientnet_b5",
        variant="b5",
        input_size=(456, 456),
        framework="tf",
        crop_pct=0.934,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b6():
    """EfficientNet-B6. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b6",
        timm_name="tf_efficientnet_b6",
        variant="b6",
        input_size=(528, 528),
        framework="tf",
        crop_pct=0.942,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b7():
    """EfficientNet-B7. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b7",
        timm_name="tf_efficientnet_b7",
        variant="b7",
        input_size=(600, 600),
        framework="tf",
        crop_pct=0.949,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b8():
    """EfficientNet-B8. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b8",
        timm_name="tf_efficientnet_b8",
        variant="b8",
        input_size=(672, 672),
        framework="tf",
        crop_pct=0.954,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b0_ap():
    """EfficientNet-B0 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b0_ap",
        timm_name="tf_efficientnet_b0_ap",
        variant="b0",
        input_size=(224, 224),
        framework="tf",
        crop_pct=0.875,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b1_ap():
    """EfficientNet-B1 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b1_ap",
        timm_name="tf_efficientnet_b1_ap",
        variant="b1",
        input_size=(240, 240),
        framework="tf",
        crop_pct=0.882,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b2_ap():
    """EfficientNet-B2 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b2_ap",
        timm_name="tf_efficientnet_b2_ap",
        variant="b2",
        input_size=(260, 260),
        framework="tf",
        crop_pct=0.890,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b3_ap():
    """EfficientNet-B3 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b3_ap",
        timm_name="tf_efficientnet_b3_ap",
        variant="b3",
        input_size=(300, 300),
        framework="tf",
        crop_pct=0.904,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b4_ap():
    """EfficientNet-B4 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b4_ap",
        timm_name="tf_efficientnet_b4_ap",
        variant="b4",
        input_size=(380, 380),
        framework="tf",
        crop_pct=0.922,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b5_ap():
    """EfficientNet-B5 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b5_ap",
        timm_name="tf_efficientnet_b5_ap",
        variant="b5",
        input_size=(456, 456),
        framework="tf",
        crop_pct=0.934,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b6_ap():
    """EfficientNet-B6 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b6_ap",
        timm_name="tf_efficientnet_b6_ap",
        variant="b6",
        input_size=(528, 528),
        framework="tf",
        crop_pct=0.942,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b7_ap():
    """EfficientNet-B7 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b7_ap",
        timm_name="tf_efficientnet_b7_ap",
        variant="b7",
        input_size=(600, 600),
        framework="tf",
        crop_pct=0.949,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b8_ap():
    """EfficientNet-B8 AdvProp. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b8_ap",
        timm_name="tf_efficientnet_b8_ap",
        variant="b8",
        input_size=(672, 672),
        framework="tf",
        crop_pct=0.954,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b0_ns():
    """EfficientNet-B0 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b0_ns",
        timm_name="tf_efficientnet_b0_ns",
        variant="b0",
        input_size=(224, 224),
        framework="tf",
        crop_pct=0.875,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b1_ns():
    """EfficientNet-B1 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b1_ns",
        timm_name="tf_efficientnet_b1_ns",
        variant="b1",
        input_size=(240, 240),
        framework="tf",
        crop_pct=0.882,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b2_ns():
    """EfficientNet-B2 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b2_ns",
        timm_name="tf_efficientnet_b2_ns",
        variant="b2",
        input_size=(260, 260),
        framework="tf",
        crop_pct=0.890,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b3_ns():
    """EfficientNet-B3 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b3_ns",
        timm_name="tf_efficientnet_b3_ns",
        variant="b3",
        input_size=(300, 300),
        framework="tf",
        crop_pct=0.904,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b4_ns():
    """EfficientNet-B4 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b4_ns",
        timm_name="tf_efficientnet_b4_ns",
        variant="b4",
        input_size=(380, 380),
        framework="tf",
        crop_pct=0.922,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b5_ns():
    """EfficientNet-B5 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b5_ns",
        timm_name="tf_efficientnet_b5_ns",
        variant="b5",
        input_size=(456, 456),
        framework="tf",
        crop_pct=0.934,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b6_ns():
    """EfficientNet-B6 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b6_ns",
        timm_name="tf_efficientnet_b6_ns",
        variant="b6",
        input_size=(528, 528),
        framework="tf",
        crop_pct=0.942,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_b7_ns():
    """EfficientNet-B7 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_b7_ns",
        timm_name="tf_efficientnet_b7_ns",
        variant="b7",
        input_size=(600, 600),
        framework="tf",
        crop_pct=0.949,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_l2_ns_475():
    """EfficientNet-L2 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_l2_ns_475",
        timm_name="tf_efficientnet_l2_ns_475",
        variant="l2",
        input_size=(475, 475),
        framework="tf",
        crop_pct=0.936,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_l2_ns():
    """EfficientNet-L2 NoisyStudent. Tensorflow compatible variant."""
    cfg = _efficientnet_cfg(
        name="efficientnet_l2_ns",
        timm_name="tf_efficientnet_l2_ns",
        variant="l2",
        input_size=(800, 800),
        framework="tf",
        crop_pct=0.96,
    )
    return EfficientNet, cfg


@register_model
def pt_efficientnet_b0():
    """EfficientNet-B0. Default version from TIMM."""
    cfg = _efficientnet_cfg(
        name="pt_efficientnet_b0",
        timm_name="efficientnet_b0",
        variant="b0",
        input_size=(224, 224),
        framework="pytorch",
        crop_pct=0.875,
    )
    return EfficientNet, cfg


@register_model
def pt_efficientnet_b1():
    """EfficientNet-B1. Default version from TIMM."""
    cfg = _efficientnet_cfg(
        name="pt_efficientnet_b1",
        timm_name="efficientnet_b1",
        variant="b1",
        input_size=(256, 256),
        framework="pytorch",
        crop_pct=1.0,
    )
    return EfficientNet, cfg


@register_model
def pt_efficientnet_b2():
    """EfficientNet-B2. Default version from TIMM."""
    cfg = _efficientnet_cfg(
        name="pt_efficientnet_b2",
        timm_name="efficientnet_b2",
        variant="b2",
        input_size=(256, 256),
        framework="pytorch",
        crop_pct=1.0,
    )
    return EfficientNet, cfg


@register_model
def pt_efficientnet_b3():
    """EfficientNet-B3. Default version from TIMM."""
    cfg = _efficientnet_cfg(
        name="pt_efficientnet_b3",
        timm_name="efficientnet_b3",
        variant="b3",
        input_size=(288, 288),
        framework="pytorch",
        crop_pct=1.0,
    )
    return EfficientNet, cfg


@register_model
def pt_efficientnet_b4():
    """EfficientNet-B4. Default version from TIMM."""
    cfg = _efficientnet_cfg(
        name="pt_efficientnet_b4",
        timm_name="efficientnet_b4",
        variant="b4",
        input_size=(320, 320),
        framework="pytorch",
        crop_pct=1.0,
    )
    return EfficientNet, cfg


def _efficientnet_edge_cfg(
    name: str,
    timm_name: str,
    variant: str,
    input_size: Tuple[int, int],
    framework: str,
    crop_pct: float,
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN,
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD,
):
    """
    Creates the config for an EfficientNet model.

    Ref impl: https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/efficientnet_model.py  # noqa: E501
    Paper: https://arxiv.org/abs/1905.11946

    Args:
        name: Model name
        timm_name: Model name in TIMM
        variant: Model variant, e.g., "b0", etc.
        framework: "tf" or "pytorch" for BN params and padding
        crop_pct: Crop percentage for ImageNet evaluation
    """
    assert framework in {"tf", "pytorch"}

    param_dict = {
        "es": (1.0, 1.0, 0.2),
        "em": (1.0, 1.1, 0.2),
        "el": (1.2, 1.4, 0.3),
    }
    channel_multiplier, depth_multiplier, drop_rate = param_dict[variant]

    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        input_size=input_size,
        stem_size=round_channels(32, multiplier=channel_multiplier),
        architecture=(
            # Note: "fc" is present to override a mismatch between stem channels and
            # in channels not present in other models
            ("er_r1_k3_s1_e4_c24_fc24_noskip",),
            ("er_r2_k3_s2_e8_c32",),
            ("er_r4_k3_s2_e8_c48",),
            ("ir_r5_k5_s2_e8_c96",),
            ("ir_r4_k5_s1_e8_c144",),
            ("ir_r2_k5_s2_e8_c192",),
        ),
        channel_multiplier=channel_multiplier,
        depth_multiplier=depth_multiplier,
        nb_features=round_channels(1280, multiplier=channel_multiplier),
        drop_rate=drop_rate,
        drop_path_rate=drop_rate,
        norm_layer="batch_norm_tf" if framework == "tf" else "batch_norm",
        act_layer="relu",
        padding="same" if framework == "tf" else "symmetric",
        crop_pct=crop_pct,
        mean=mean,
        std=std,
    )
    return cfg


@register_model
def efficientnet_es():
    """EfficientNet-Edge Small. Tensorflow compatible variant."""
    cfg = _efficientnet_edge_cfg(
        name="efficientnet_es",
        timm_name="tf_efficientnet_es",
        variant="es",
        input_size=(224, 224),
        framework="tf",
        crop_pct=0.875,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_em():
    """EfficientNet-Edge Medium. Tensorflow compatible variant."""
    cfg = _efficientnet_edge_cfg(
        name="efficientnet_em",
        timm_name="tf_efficientnet_em",
        variant="em",
        input_size=(240, 240),
        framework="tf",
        crop_pct=0.882,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_el():
    """EfficientNet-Edge Large. Tensorflow compatible variant."""
    cfg = _efficientnet_edge_cfg(
        name="efficientnet_el",
        timm_name="tf_efficientnet_el",
        variant="el",
        input_size=(300, 300),
        framework="tf",
        crop_pct=0.904,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return EfficientNet, cfg


def _efficientnet_v2_base_cfg(
    name: str,
    timm_name: str,
    variant: str,
    input_size: Tuple[int, int],
    crop_pct: float,
):
    """
    Creates the config for an EfficientNet-V2 base model.

    Ref impl: https://github.com/google/automl/tree/master/efficientnetv2
    Paper: `EfficientNetV2: Smaller Models and Faster Training`
    Link: https://arxiv.org/abs/2104.00298
    """
    param_dict = {
        "b0": (1.0, 1.0, 0.2),
        "b1": (1.0, 1.1, 0.2),
        "b2": (1.1, 1.2, 0.3),
        "b3": (1.2, 1.4, 0.3),
    }
    channel_multiplier, depth_multiplier, drop_rate = param_dict[variant]
    round_chs_fn = partial(
        round_channels, multiplier=channel_multiplier, round_limit=0.0
    )
    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        input_size=input_size,
        stem_size=round_chs_fn(32),
        architecture=(
            ("cn_r1_k3_s1_e1_c16_skip",),
            ("er_r2_k3_s2_e4_c32",),
            ("er_r2_k3_s2_e4_c48",),
            ("ir_r3_k3_s2_e4_c96_se0.25",),
            ("ir_r5_k3_s1_e6_c112_se0.25",),
            ("ir_r8_k3_s2_e6_c192_se0.25",),
        ),
        channel_multiplier=channel_multiplier,
        depth_multiplier=depth_multiplier,
        nb_features=round_chs_fn(1280),
        drop_rate=drop_rate,
        drop_path_rate=drop_rate,
        norm_layer="batch_norm_tf",
        act_layer="swish",
        padding="same",
        crop_pct=crop_pct,
    )
    return cfg


@register_model
def efficientnet_v2_b0():
    """EfficientNet-V2-B0. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_base_cfg(
        name="efficientnet_v2_b0",
        timm_name="tf_efficientnetv2_b0",
        variant="b0",
        input_size=(192, 192),
        crop_pct=0.875,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_b1():
    """EfficientNet-V2-B1. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_base_cfg(
        name="efficientnet_v2_b1",
        timm_name="tf_efficientnetv2_b1",
        variant="b1",
        input_size=(192, 192),
        crop_pct=0.882,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_b2():
    """EfficientNet-V2-b2. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_base_cfg(
        name="efficientnet_v2_b2",
        timm_name="tf_efficientnetv2_b2",
        variant="b2",
        input_size=(208, 208),
        crop_pct=0.890,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_b3():
    """EfficientNet-V2-b3. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_base_cfg(
        name="efficientnet_v2_b3",
        timm_name="tf_efficientnetv2_b3",
        variant="b3",
        input_size=(240, 240),
        crop_pct=0.904,
    )
    return EfficientNet, cfg


def _efficientnet_v2_s_cfg(
    name: str,
    timm_name: str,
    nb_classes: int = 1000,
):
    """
    Creates the config for an EfficientNet-V2 small model.

    Ref impl: https://github.com/google/automl/tree/master/efficientnetv2
    Paper: `EfficientNetV2: Smaller Models and Faster Training`
    Link: https://arxiv.org/abs/2104.00298
    """
    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        nb_classes=nb_classes,
        input_size=(300, 300),
        stem_size=24,
        architecture=(
            ("cn_r2_k3_s1_e1_c24_skip",),
            ("er_r4_k3_s2_e4_c48",),
            ("er_r4_k3_s2_e4_c64",),
            ("ir_r6_k3_s2_e4_c128_se0.25",),
            ("ir_r9_k3_s1_e6_c160_se0.25",),
            ("ir_r15_k3_s2_e6_c256_se0.25",),
        ),
        channel_multiplier=1.0,
        depth_multiplier=1.0,
        nb_features=1280,
        drop_rate=0.3,
        drop_path_rate=0.3,
        norm_layer="batch_norm_tf",
        act_layer="swish",
        padding="same",
        crop_pct=1.0,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return cfg


@register_model
def efficientnet_v2_s():
    """EfficientNet-V2 Small. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_s_cfg(
        name="efficientnet_v2_s",
        timm_name="tf_efficientnetv2_s",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_s_in21ft1k():
    """
    EfficientNet-V2 Small. Pretrained on ImageNet-21k, fine-tuned on 1k. Tensorflow
    compatible variant.
    """
    cfg = _efficientnet_v2_s_cfg(
        name="efficientnet_v2_s_in21ft1k",
        timm_name="tf_efficientnetv2_s_in21ft1k",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_s_in21k():
    """
    EfficientNet-V2 Small. ImageNet-21k pre-trained weights. Tensorflow compatible
    variant.
    """
    cfg = _efficientnet_v2_s_cfg(
        name="efficientnet_v2_s_in21k",
        timm_name="tf_efficientnetv2_s_in21k",
        nb_classes=21843,
    )
    return EfficientNet, cfg


def _efficientnet_v2_m_cfg(
    name: str,
    timm_name: str,
    nb_classes: int = 1000,
):
    """
    Creates the config for an EfficientNet-V2 medium model.

    Ref impl: https://github.com/google/automl/tree/master/efficientnetv2
    Paper: `EfficientNetV2: Smaller Models and Faster Training`
    Link: https://arxiv.org/abs/2104.00298
    """
    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        nb_classes=nb_classes,
        input_size=(384, 384),
        stem_size=24,
        architecture=(
            ("cn_r3_k3_s1_e1_c24_skip",),
            ("er_r5_k3_s2_e4_c48",),
            ("er_r5_k3_s2_e4_c80",),
            ("ir_r7_k3_s2_e4_c160_se0.25",),
            ("ir_r14_k3_s1_e6_c176_se0.25",),
            ("ir_r18_k3_s2_e6_c304_se0.25",),
            ("ir_r5_k3_s1_e6_c512_se0.25",),
        ),
        channel_multiplier=1.0,
        depth_multiplier=1.0,
        nb_features=1280,
        drop_rate=0.4,
        drop_path_rate=0.4,
        norm_layer="batch_norm_tf",
        act_layer="swish",
        padding="same",
        crop_pct=1.0,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return cfg


@register_model
def efficientnet_v2_m():
    """EfficientNet-V2 Medium. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_m_cfg(
        name="efficientnet_v2_m",
        timm_name="tf_efficientnetv2_m",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_m_in21ft1k():
    """
    EfficientNet-V2 Medium. Pretrained on ImageNet-21k, fine-tuned on 1k. Tensorflow
    compatible variant.
    """
    cfg = _efficientnet_v2_m_cfg(
        name="efficientnet_v2_m_in21ft1k",
        timm_name="tf_efficientnetv2_m_in21ft1k",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_m_in21k():
    """
    EfficientNet-V2 Medium. ImageNet-21k pre-trained weights. Tensorflow compatible
    variant.
    """
    cfg = _efficientnet_v2_m_cfg(
        name="efficientnet_v2_m_in21k",
        timm_name="tf_efficientnetv2_m_in21k",
        nb_classes=21843,
    )
    return EfficientNet, cfg


def _efficientnet_v2_l_cfg(
    name: str,
    timm_name: str,
    nb_classes: int = 1000,
):
    """
    Creates the config for an EfficientNet-V2 large model.

    Ref impl: https://github.com/google/automl/tree/master/efficientnetv2
    Paper: `EfficientNetV2: Smaller Models and Faster Training`
    Link: https://arxiv.org/abs/2104.00298
    """
    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        nb_classes=nb_classes,
        input_size=(384, 384),
        stem_size=32,
        architecture=(
            ("cn_r4_k3_s1_e1_c32_skip",),
            ("er_r7_k3_s2_e4_c64",),
            ("er_r7_k3_s2_e4_c96",),
            ("ir_r10_k3_s2_e4_c192_se0.25",),
            ("ir_r19_k3_s1_e6_c224_se0.25",),
            ("ir_r25_k3_s2_e6_c384_se0.25",),
            ("ir_r7_k3_s1_e6_c640_se0.25",),
        ),
        channel_multiplier=1.0,
        depth_multiplier=1.0,
        nb_features=1280,
        drop_rate=0.5,
        drop_path_rate=0.5,
        norm_layer="batch_norm_tf",
        act_layer="swish",
        padding="same",
        crop_pct=1.0,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return cfg


@register_model
def efficientnet_v2_l():
    """EfficientNet-V2 Large. Tensorflow compatible variant."""
    cfg = _efficientnet_v2_l_cfg(
        name="efficientnet_v2_l",
        timm_name="tf_efficientnetv2_l",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_l_in21ft1k():
    """
    EfficientNet-V2 Large. Pretrained on ImageNet-21k, fine-tuned on 1k. Tensorflow
    compatible variant.
    """
    cfg = _efficientnet_v2_l_cfg(
        name="efficientnet_v2_l_in21ft1k",
        timm_name="tf_efficientnetv2_l_in21ft1k",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_l_in21k():
    """
    EfficientNet-V2 Large. ImageNet-21k pre-trained weights. Tensorflow compatible
    variant.
    """
    cfg = _efficientnet_v2_l_cfg(
        name="efficientnet_v2_l_in21k",
        timm_name="tf_efficientnetv2_l_in21k",
        nb_classes=21843,
    )
    return EfficientNet, cfg


def _efficientnet_v2_xl_cfg(
    name: str,
    timm_name: str,
    nb_classes: int = 1000,
):
    """
    Creates the config for an EfficientNet-V2 XLarge model.

    Ref impl: https://github.com/google/automl/tree/master/efficientnetv2
    Paper: `EfficientNetV2: Smaller Models and Faster Training`
    Link: https://arxiv.org/abs/2104.00298
    """
    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        nb_classes=nb_classes,
        input_size=(384, 384),
        stem_size=32,
        architecture=(
            ("cn_r4_k3_s1_e1_c32_skip",),
            ("er_r8_k3_s2_e4_c64",),
            ("er_r8_k3_s2_e4_c96",),
            ("ir_r16_k3_s2_e4_c192_se0.25",),
            ("ir_r24_k3_s1_e6_c256_se0.25",),
            ("ir_r32_k3_s2_e6_c512_se0.25",),
            ("ir_r8_k3_s1_e6_c640_se0.25",),
        ),
        channel_multiplier=1.0,
        depth_multiplier=1.0,
        nb_features=1280,
        drop_rate=0.5,
        drop_path_rate=0.5,
        norm_layer="batch_norm_tf",
        act_layer="swish",
        padding="same",
        crop_pct=1.0,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return cfg


@register_model
def efficientnet_v2_xl_in21ft1k():
    """
    EfficientNet-V2 XLarge. Pretrained on ImageNet-21k, fine-tuned on 1k. Tensorflow
    compatible variant.
    """
    cfg = _efficientnet_v2_xl_cfg(
        name="efficientnet_v2_xl_in21ft1k",
        timm_name="tf_efficientnetv2_xl_in21ft1k",
    )
    return EfficientNet, cfg


@register_model
def efficientnet_v2_xl_in21k():
    """
    EfficientNet-V2 XLarge. ImageNet-21k pre-trained weights. Tensorflow compatible
    variant.
    """
    cfg = _efficientnet_v2_xl_cfg(
        name="efficientnet_v2_xl_in21k",
        timm_name="tf_efficientnetv2_xl_in21k",
        nb_classes=21843,
    )
    return EfficientNet, cfg


def _efficientnet_lite_cfg(
    name: str,
    timm_name: str,
    variant: str,
    crop_pct: float,
):
    """
    Creates the config for an EfficientNet model.

    Ref impl: https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/efficientnet_model.py  # noqa: E501
    Paper: https://arxiv.org/abs/1905.11946

    Args:
        name: Model name
        timm_name: Model name in TIMM
        variant: Model variant, e.g., "lite0", etc.
        crop_pct: Crop percentage for ImageNet evaluation
    """
    param_dict = {
        "lite0": (1.0, 1.0, 224, 0.2),
        "lite1": (1.0, 1.1, 240, 0.2),
        "lite2": (1.1, 1.2, 260, 0.3),
        "lite3": (1.2, 1.4, 280, 0.3),
        "lite4": (1.4, 1.8, 300, 0.3),
    }
    channel_multiplier, depth_multiplier, input_size, drop_rate = param_dict[variant]
    input_size = (input_size, input_size)

    cfg = EfficientNetConfig(
        name=name,
        url="[timm]" + timm_name,
        input_size=input_size,
        stem_size=32,
        architecture=(
            ("ds_r1_k3_s1_e1_c16",),
            ("ir_r2_k3_s2_e6_c24",),
            ("ir_r2_k5_s2_e6_c40",),
            ("ir_r3_k3_s2_e6_c80",),
            ("ir_r3_k5_s1_e6_c112",),
            ("ir_r4_k5_s2_e6_c192",),
            ("ir_r1_k3_s1_e6_c320",),
        ),
        channel_multiplier=channel_multiplier,
        depth_multiplier=depth_multiplier,
        fix_first_last=True,
        nb_features=1280,
        drop_rate=drop_rate,
        drop_path_rate=drop_rate,
        norm_layer="batch_norm_tf",
        act_layer="relu6",
        padding="same",
        crop_pct=crop_pct,
        mean=IMAGENET_INCEPTION_MEAN,
        std=IMAGENET_INCEPTION_STD,
    )
    return cfg


@register_model
def efficientnet_lite0():
    """EfficientNet-Lite0. Tensorflow compatible variant."""
    cfg = _efficientnet_lite_cfg(
        name="efficientnet_lite0",
        timm_name="tf_efficientnet_lite0",
        variant="lite0",
        crop_pct=0.875,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_lite1():
    """EfficientNet-Lite1. Tensorflow compatible variant."""
    cfg = _efficientnet_lite_cfg(
        name="efficientnet_lite1",
        timm_name="tf_efficientnet_lite1",
        variant="lite1",
        crop_pct=0.882,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_lite2():
    """EfficientNet-Lite2. Tensorflow compatible variant."""
    cfg = _efficientnet_lite_cfg(
        name="efficientnet_lite2",
        timm_name="tf_efficientnet_lite2",
        variant="lite2",
        crop_pct=0.890,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_lite3():
    """EfficientNet-Lite3. Tensorflow compatible variant."""
    cfg = _efficientnet_lite_cfg(
        name="efficientnet_lite3",
        timm_name="tf_efficientnet_lite3",
        variant="lite3",
        crop_pct=0.904,
    )
    return EfficientNet, cfg


@register_model
def efficientnet_lite4():
    """EfficientNet-Lite4. Tensorflow compatible variant."""
    cfg = _efficientnet_lite_cfg(
        name="efficientnet_lite4",
        timm_name="tf_efficientnet_lite4",
        variant="lite4",
        crop_pct=0.920,
    )
    return EfficientNet, cfg



import collections.abc
from typing import Iterable, Optional, Tuple, TypeVar, Union

T = TypeVar("T")


def to_2tuple(x: Union[T, Iterable[T]]) -> Tuple[T, T]:
    if isinstance(x, collections.abc.Iterable):
        return tuple(x)[:2]
    else:
        return x, x


def make_divisible(
    value: float,
    divisor: int,
    min_value: Optional[int] = None,
    round_limit: float = 0.9,
) -> int:
    """Ensures that `value` is divisible by `divisor`."""
    min_value = min_value or divisor
    new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_value < round_limit * value:
        new_value += divisor
    return new_value



from abc import ABC, abstractmethod


class ProblemBase(ABC):
    """Problem base class describing the interface expected from each problem class."""

    @abstractmethod
    def train_step(self, data, it: int) -> (float, dict):
        """
        Perform one step of training. Each problem class has to implement at least
        this method. The function should return a loss, used to display training
        progress and a dictionary of metrics that will be logged to W&B.
        """
        raise NotImplementedError

    def start_epoch(self):
        """Called at the beginning of an epoch. Can be used to reset Keras metrics."""
        pass

    def ckpt_variables(self, model_only: bool = False) -> dict:
        """
        Return dictionary with all variables that need to be added to checkpoint.

        If `model_only=True`, return only the model variables, i.e., those variables
        that we would want to transfer across training scripts. This should exclude
        optimizer state or running averages.
        """
        return {}

    def validation(self, dataset) -> (float, dict):
        """
        Function performs validation on a dataset and returns a dictionary of metrics.
        """
        return 0.0, {}

    def save_model(self, save_dir):
        """
        The trainer class saves models as checkpoints. We may want to save them in a
        more usable form, e.g., as Keras models or saved_models. This saving has to
        be problem specific.
        """
        pass
